"""Checks for various circumstances."""

from __future__ import annotations

import asyncio
from abc import ABC, abstractmethod
from collections import defaultdict
from logging import debug, info
from pathlib import Path

from pydantic import HttpUrl

from repod.common.enums import ActionStateEnum, ArchitectureEnum, PkgTypeEnum
from repod.config.settings import UrlValidationSettings
from repod.errors import RepoManagementFileError
from repod.files import Package
from repod.files.pkginfo import PkgInfoV2, PkgType
from repod.repo.management import OutputPackageBase
from repod.repo.package.repofile import filename_parts
from repod.verification import PacmanKeyVerifier
from repod.version.alpm import pkg_vercmp


class Check(ABC):
    """An abstract base class to describe a check.

    Checks are Callables, that run a check (e.g. on an input) and must not alter data.

    Attributes
    ----------
    state: ActionStateEnum
        A member of ActionStateEnum indicating whether a Check is not started, started, failed or succeeded (defaults to
        ActionStateEnum.NOT_STARTED
    """

    state: ActionStateEnum = ActionStateEnum.NOT_STARTED

    @abstractmethod
    def __call__(self) -> ActionStateEnum:  # pragma: no cover
        """Call a Check.

        The method is expected to set the Check's state property to ActionStateEnum.STARTED, run a check
        operation, set the state property to either ActionStateEnum.SUCCESS or to ActionStateEnum.FAILED (depending on
        whether the check operation finished successfully or not, respectively) and return the state.

        Returns
        -------
        ActionStateEnum
            ActionStateEnum.SUCCESS if the check passed successfully,
            ActionStateEnum.FAILED otherwise
        """
        pass


class PacmanKeyPackagesSignatureVerificationCheck(Check):
    """Verify a list of package signatures using pacman-key.

    This Check fails if any package can not be verified with its corresponding signature or if any of the
    package/signature lists is not of length two.

    Attributes
    ----------
    packages: list[list[Path]]
        A list of Path lists, that should contain a package and a corresponding signature Path each
    """

    def __init__(self, packages: list[list[Path]]) -> None:
        """Initialize an instance of PacmanKeyPackagesSignatureVerificationCheck.

        Parameters
        ----------
        package: list[list[Path]]
            A list of lists, containing up to two Paths each
        """
        self.packages = packages

    def __call__(self) -> ActionStateEnum:
        """Use an instance of PacmanKeyVerifier to verify a package and its signature.

        Returns
        -------
        ActionStateEnum
            ActionStateEnum.SUCCESS if the check passed successfully,
            ActionStateEnum.FAILED otherwise
        """
        self.state = ActionStateEnum.STARTED

        if not all(len(package_list) == 2 for package_list in self.packages):
            info("Verification with pacman-key is requested, but not all packages provide a signature!")
            self.state = ActionStateEnum.FAILED
            return self.state

        verifier = PacmanKeyVerifier()
        debug(f"Verifying list of packages using pacman-key: {self.packages}")
        for package_list in self.packages:
            if verifier.verify(
                package=package_list[0],
                signature=package_list[1],
            ):
                debug("Package {package_list[0]} successfully verified using signature {package_list[1]}!")
            else:
                info(
                    f"Verification of package {package_list[0]} with signature {package_list[1]} "
                    "failed using pacman-key!"
                )
                self.state = ActionStateEnum.FAILED
                return self.state

        self.state = ActionStateEnum.SUCCESS
        return self.state


class DebugPackagesCheck(Check):
    """A Check to evaluate whether all instances in a list of packages are either debug or not debug packages.

    This Check currently only works successfully with packages that use PkgInfoV2, as before that PkgInfo
    implementation, figuring out whether a package is a debug package is merely guess work.
    Package files providing a PkgInfo version lower than PkgInfoV2 will always pass this test!

    Attributes
    ----------
    packages: list[Package]
        A list of Package instances to check
    debug: bool
        A boolean value indicating whether all Package instances should be debug packages or not
    """

    def __init__(self, packages: list[Package], debug: bool) -> None:
        """Initialize an instance of DebugPackagesCheck.

        Parameters
        ----------
        packages: list[Package]
            A list of Package instances to check
        debug: bool
            A boolean value indicating whether all Package instances should be either debug packages or not
        """
        self.packages = packages
        self.debug = debug

    def __call__(self) -> ActionStateEnum:
        """Check whether all instances of self.packages are supposed to be debug packages or not.

        Returns
        -------
        ActionStateEnum
            ActionStateEnum.SUCCESS if the check is successful,
            ActionStateEnum.FAILED otherwise
        """
        self.state = ActionStateEnum.STARTED

        debug(f"Testing whether all packages are either default or debug packages: {self.packages}")
        if self.debug:
            if any(
                [
                    True
                    for package in self.packages
                    if isinstance(
                        package.pkginfo,  # type: ignore[attr-defined]
                        PkgInfoV2,
                    )
                    and PkgType(pkgtype=PkgTypeEnum.DEBUG) not in package.pkginfo.xdata  # type: ignore[attr-defined]
                ]
            ):
                info("A debug repository is targetted, but not all provided packages are debug packages!")
                self.state = ActionStateEnum.FAILED
                return self.state
        else:
            if any(
                [
                    True
                    for package in self.packages
                    if isinstance(
                        package.pkginfo,  # type: ignore[attr-defined]
                        PkgInfoV2,
                    )
                    and PkgType(pkgtype=PkgTypeEnum.DEBUG) in package.pkginfo.xdata  # type: ignore[attr-defined]
                ]
            ):
                info("A non-debug repository is targetted, but not all provided packages are non-debug packages!")
                self.state = ActionStateEnum.FAILED
                return self.state

        self.state = ActionStateEnum.SUCCESS
        return self.state


class MatchingArchitectureCheck(Check):
    """A Check to ensure that a list of packages match a target CPU architecture.

    Attributes
    ----------
    architecture: ArchitectureEnum
        An instance of ArchitectureEnum which identifies the target CPU architecture
    packages: list[Package]
        A list of Package instances to check
    """

    def __init__(self, architecture: ArchitectureEnum, packages: list[Package]) -> None:
        """Initialize an instance of MatchingArchitectureCheck.

        Parameters
        ----------
        architecture: ArchitectureEnum
            An instance of ArchitectureEnum which identifies the target CPU architecture
        packages: list[Package]
            A list of Package instances to check
        """
        self.architecture = architecture
        self.packages = packages

    def __call__(self) -> ActionStateEnum:
        """Check whether all instances of self.packages are supposed to be debug packages or not.

        Returns
        -------
        ActionStateEnum
            ActionStateEnum.SUCCESS if the check is successful,
            ActionStateEnum.FAILED otherwise
        """
        self.state = ActionStateEnum.STARTED

        debug("Running check to test whether all packages match the target architecture...")

        non_matching: list[str] = []

        for package in self.packages:
            if (
                package.pkginfo.arch != self.architecture.value  # type: ignore[attr-defined]
                and package.pkginfo.arch != ArchitectureEnum.ANY.value  # type: ignore[attr-defined]
            ):
                non_matching.append(f"{package.pkginfo.name}/{package.pkginfo.arch}")  # type: ignore[attr-defined]

        if len(non_matching) > 0:
            self.state = ActionStateEnum.FAILED
            info(
                "Adding package(s) to repository failed as the following packages are not compatible with CPU "
                f"architecture {'(' + self.architecture.value + ')' if self.architecture else ''}: {non_matching}"
            )
        else:
            self.state = ActionStateEnum.SUCCESS

        return self.state


class MatchingFilenameCheck(Check):
    """A Check to ensure that package metadata matches the data from the package filename.

    Attributes
    ----------
    packages_and_paths: list[tuple[Package, Path]]
        A list of tuples of Package instances and Path instances to check
    """

    def __init__(self, packages_and_paths: list[tuple[Package, Path]]) -> None:
        """Initialize an instance of MatchingFilenameCheck.

        Parameters
        ----------
        packages_and_paths: list[tuple[Package, Path]]
            A list of tuples of Package instances and Path instances to check
        """
        self.packages_and_paths = packages_and_paths

    def __call__(self) -> ActionStateEnum:
        """Check that package metadata matches data from the filename.

        Returns
        -------
        ActionStateEnum
            ActionStateEnum.SUCCESS if the check is successful,
            ActionStateEnum.FAILED otherwise
        """
        self.state = ActionStateEnum.STARTED

        debug("Running check to test whether package metadata and package filenames match...")

        non_matching: list[str] = []

        for package, path in self.packages_and_paths:
            pkginfo = package.pkginfo  # type: ignore[attr-defined]
            filename_data = filename_parts(path)
            pkg_data = {"arch": pkginfo.arch, "name": pkginfo.name, "version": pkginfo.version}
            for key, value in pkg_data.items():
                if filename_data[key] != value:
                    msg = f"{package.filename}: {filename_data[key]}/{value}"  # type: ignore[attr-defined]
                    non_matching.append(msg)

        if len(non_matching) > 0:
            self.state = ActionStateEnum.FAILED
            info(
                "Adding package(s) to repository failed as the following packages have a mismatch between package "
                f"filename and package metadata: {non_matching}"
            )
        else:
            self.state = ActionStateEnum.SUCCESS

        return self.state


class PkgbasesVersionUpdateCheck(Check):
    """A Check to ensure, that pkgbases are updated, not downgraded.

    Attributes
    ----------
    new_pkgbases: list[OutputPackageBase]
        A list of updated OutputPackageBase instances
    current_pkgbases: list[OutputPackageBase]
        A list of current OutputPackageBase instances
    """

    def __init__(
        self,
        new_pkgbases: list[OutputPackageBase],
        current_pkgbases: list[OutputPackageBase],
    ) -> None:
        """Initialize an instance of PkgbasesVersionUpdateCheck.

        Parameters
        ----------
        new_pkgbases: list[OutputPackageBase]
            A list of updated OutputPackageBase instances
        current_pkgbases: list[OutputPackageBase]
            A list of current OutputPackageBase instances
        """
        self.new_pkgbases = new_pkgbases
        self.current_pkgbases = current_pkgbases

    def __call__(self) -> ActionStateEnum:
        """Check, that pkgbases are updated, not downgraded.

        Returns
        -------
        ActionStateEnum
            ActionStateEnum.SUCCESS if the check passed successfully,
            ActionStateEnum.FAILED otherwise
        """
        self.state = ActionStateEnum.STARTED

        debug("Running check to test whether all pkgbases are being upgraded, not downgraded...")

        current_names_versions = [
            {"name": c_pkgbase.base, "version": c_pkgbase.version}  # type: ignore[attr-defined]
            for c_pkgbase in self.current_pkgbases
        ]

        for pkgbase in self.new_pkgbases:
            name = pkgbase.base  # type: ignore[attr-defined]
            version = pkgbase.version  # type: ignore[attr-defined]
            if name in [c_dict.get("name") for c_dict in current_names_versions]:
                c_version = [
                    str(c_dict.get("version")) for c_dict in current_names_versions if str(c_dict.get("name")) == name
                ][0]
                if (
                    pkg_vercmp(
                        c_version,
                        version,
                    )
                    >= 0
                ):
                    info(
                        f"The version of {name} currently "
                        f"in the repository is newer than the provided one: {c_version} vs. {version}"
                    )
                    self.state = ActionStateEnum.FAILED
                    return self.state

        self.state = ActionStateEnum.SUCCESS
        return self.state


class PackagesNewOrUpdatedCheck(Check):
    """A Check to ensure, that packages are new or updated.

    Attributes
    ----------
    directory: Path
        A Path to the directory in a management repository where to look for OutputPackageBases
    new_pkgbases: list[OutputPackageBase]
        A list of updated OutputPackageBase instances
    current_pkgbases: list[OutputPackageBase]
        A list of current OutputPackageBase instances
    """

    def __init__(
        self,
        directory: Path,
        new_pkgbases: list[OutputPackageBase],
        current_pkgbases: list[OutputPackageBase],
    ) -> None:
        """Initialize an instance of PackagesNewOrUpdatedCheck.

        Parameters
        ----------
        directory: Path
            A Path to the directory in a management repository where to look for OutputPackageBases
        new_pkgbases: list[OutputPackageBase]
            A list of updated OutputPackageBase instances
        current_pkgbases: list[OutputPackageBase]
            A list of current OutputPackageBase instances
        """
        self.directory = directory
        self.new_pkgbases = new_pkgbases
        self.current_pkgbases = current_pkgbases

    def __call__(self) -> ActionStateEnum:
        """Check, that packages are new or updated.

        Returns
        -------
        ActionStateEnum
            ActionStateEnum.SUCCESS if the check passed successfully,
            ActionStateEnum.FAILED otherwise
        """
        self.state = ActionStateEnum.STARTED

        debug("Running check to test that all packages are either new or updated...")

        current_pkgbases = [
            {
                "name": c_pkgbase.base,  # type: ignore[attr-defined]
                "packages": [pkg.name for pkg in c_pkgbase.packages],  # type: ignore[attr-defined]
                "version": c_pkgbase.version,  # type: ignore[attr-defined]
            }
            for c_pkgbase in self.current_pkgbases
        ]
        new_pkgbases = [
            {
                "name": pkgbase.base,  # type: ignore[attr-defined]
                "packages": [pkg.name for pkg in pkgbase.packages],  # type: ignore[attr-defined]
                "version": pkgbase.version,  # type: ignore[attr-defined]
            }
            for pkgbase in self.new_pkgbases
        ]

        for pkgbase in new_pkgbases:
            target_pkgbase_file = self.directory / Path(str(pkgbase.get("name")) + ".json")

            for package in pkgbase.get("packages", []):
                target_package_file = self.directory / "pkgnames" / Path(package + ".json")
                target_package_pkgbase = target_package_file.resolve().name.replace(".json", "")
                original_pkgbase_in_new_pkgbases = any(
                    True for pkgbase in new_pkgbases if pkgbase.get("name") == target_package_pkgbase
                )
                original_pkgbase_provides_package = any(
                    [
                        True
                        for pkgbase in new_pkgbases
                        if pkgbase.get("name") == target_package_pkgbase and package in str(pkgbase.get("packages"))
                    ]
                )

                # the pkgbase of the new package does not match an existing pkgbase (in current_pkgbases or
                # new_pkgbases), but a file exists and provides a version newer than the one added
                if (
                    target_package_file.is_symlink()
                    and target_package_file.resolve() != target_pkgbase_file
                    and pkgbase.get("name") not in [pkgbase.get("name") for pkgbase in current_pkgbases]
                ):
                    try:
                        old_pkgbase = asyncio.run(OutputPackageBase.from_file(target_package_file.resolve()))
                    except RepoManagementFileError as e:
                        info(e)
                        self.state = ActionStateEnum.FAILED
                        return self.state

                    old_version = old_pkgbase.get_version()
                    new_version = str(pkgbase.get("version"))

                    if pkg_vercmp(old_version, new_version) >= 0:
                        info(
                            f"The version of the added {package} (provided by pkgbase {pkgbase.get('name')}) "
                            "is newer or equal to the one already in the repository (provided by pkgbase "
                            f"{target_package_pkgbase}): {old_version} (old) vs. {new_version} (new)"
                        )
                        self.state = ActionStateEnum.FAILED
                        return self.state

                # the pkgbase of the new package does not match an existing pkgbase and the update also does not
                # remove the package from the previous pkgbase
                if (
                    target_package_file.is_symlink()
                    and target_package_file.resolve() != target_pkgbase_file
                    and (
                        not original_pkgbase_in_new_pkgbases
                        or (original_pkgbase_in_new_pkgbases and original_pkgbase_provides_package)
                    )
                ):
                    info(
                        f"The package {package} is currently provided by "
                        f"pkgbase {target_package_pkgbase}, but the new pkgbase "
                        f"{pkgbase.get('name')} now tries to provide it, "
                        f"without removing the package from the pkgbase {target_package_pkgbase}."
                    )
                    self.state = ActionStateEnum.FAILED
                    return self.state

        self.state = ActionStateEnum.SUCCESS
        return self.state


class SourceUrlCheck(Check):
    """Check that pkgbases have a source url, if their repository validation settings require it.

    Attributes
    ----------
    new_pkgbases: dict[str, HttpUrl | None]
        A dict representing new pkgbases and their respective optional upstream URL (to check)
    current_pkgbase_urls: dict[str, HttpUrl | None]
        A dict representing current pkgbases and their respective optional upstream URL
    url_validation_settings: UrlValidationSettings | None
        An optional UrlValidationSettings instance with which to validate source URLs of the provided new_pkgbases
    """

    def __init__(
        self,
        new_pkgbases: list[OutputPackageBase],
        current_pkgbases: list[OutputPackageBase],
        url_validation_settings: UrlValidationSettings | None,
    ) -> None:
        """Initialize an instance of SourceUrlCheck.

        Parameters
        ----------
        new_pkgbases: list[OutputPackageBase]
            A list of OutputPackageBase instances representing new pkgbases to check
        current_pkgbases: list[OutputPackageBase]
            A list of current OutputPackageBase instances (a subset of new_pkgbases)
        url_validation_settings: UrlValidationSettings | None
            An optional UrlValidationSettings instance with which to validate source URLs of the provided new_pkgbases
        """
        self.new_pkgbase_urls: dict[str, HttpUrl | None] = {
            pkgbase.base: pkgbase.source_url for pkgbase in new_pkgbases  # type: ignore[attr-defined]
        }
        self.current_pkgbase_urls: dict[str, HttpUrl | None] = {
            pkgbase.base: pkgbase.source_url for pkgbase in current_pkgbases  # type: ignore[attr-defined]
        }
        self.url_validation_settings = url_validation_settings

    def __call__(self) -> ActionStateEnum:
        """Check that pkgbases have a source url, if their repository validation settings require it.

        Returns
        -------
        ActionStateEnum
            ActionStateEnum.SUCCESS if the check passed successfully,
            ActionStateEnum.FAILED otherwise
        """
        self.state = ActionStateEnum.STARTED

        debug("Running check to validate the URLs of pkgbases... ")

        if not self.url_validation_settings:
            debug("No URL validation required, skipping check...")
            self.state = ActionStateEnum.SUCCESS
            return self.state

        for pkgbase, url in self.new_pkgbase_urls.items():
            url = url or self.current_pkgbase_urls.get(pkgbase)

            if not url:
                info(f"The pkgbase {pkgbase} does neither already have a source URL set nor is one provided for it!")
                self.state = ActionStateEnum.FAILED
                return self.state

            if not self.url_validation_settings.validate_url(url=url):
                info(
                    f"The source URL of the pkgbase {pkgbase} ({url}) does not validate "
                    "against the repository's settings!"
                )
                self.state = ActionStateEnum.FAILED
                return self.state

        self.state = ActionStateEnum.SUCCESS
        return self.state


class StabilityLayerCheck(Check):
    """A Check to compare the versions of pkgbases in the stability layers above and below them.

    Attributes
    ----------
    pkgbases: list[OutputPackageBase]
        A list of OutputPackageBase objects, that represent the pkgbases to be checked
    pkgbases_above: list[OutputPackageBase]
        A list of OutputPackageBase objects, that represent the pkgbases in stability layers above those in pkgbases
    pkgbases_below: list[OutputPackageBase]
        A list of OutputPackageBase objects, that represent the pkgbases in stability layers below those in pkgbases
    """

    def __init__(
        self,
        pkgbases: list[OutputPackageBase],
        pkgbases_above: list[OutputPackageBase],
        pkgbases_below: list[OutputPackageBase],
    ) -> None:
        """Initialize an instance of StabilityLayerCheck.

        Parameters
        ----------
        pkgbases: list[OutputPackageBase]
            A list of OutputPackageBase objects, that represent the pkgbases to be checked
        pkgbases_above: list[OutputPackageBase]
            A list of OutputPackageBase objects, that represent the pkgbases in stability layers above those in pkgbases
        pkgbases_below: list[OutputPackageBase]
            A list of OutputPackageBase objects, that represent the pkgbases in stability layers below those in pkgbases
        """
        self.pkgbases = pkgbases
        self.pkgbases_above = pkgbases_above
        self.pkgbases_below = pkgbases_below

    def __call__(self) -> ActionStateEnum:
        """Check that pkgbases are not newer than pkgbases_above and not older than pkgbases_below.

        Returns
        -------
        ActionStateEnum
            ActionStateEnum.SUCCESS if the check passed successfully,
            ActionStateEnum.FAILED otherwise
        """
        self.state = ActionStateEnum.STARTED

        for pkgbase in self.pkgbases:
            name = pkgbase.base  # type: ignore[attr-defined]
            version = pkgbase.get_version()
            above_versions = [
                pkgbase_above.get_version()
                for pkgbase_above in self.pkgbases_above
                if pkgbase_above.base == pkgbase.base  # type: ignore[attr-defined]
            ]
            for above_version in above_versions:
                if pkg_vercmp(a=version, b=above_version) >= 0:
                    info(
                        f"The version of {name} ({version}) is newer than what is available "
                        f"in a stability layer above: {above_version}"
                    )
                    self.state = ActionStateEnum.FAILED
                    return self.state

            below_versions = [
                pkgbase_below.get_version()
                for pkgbase_below in self.pkgbases_below
                if pkgbase_below.base == pkgbase.base  # type: ignore[attr-defined]
            ]
            for below_version in below_versions:
                if pkg_vercmp(a=version, b=below_version) <= 0:
                    info(
                        f"The version of {name} ({version}) is older than what is available "
                        f"in a stability layer below: {below_version}"
                    )
                    self.state = ActionStateEnum.FAILED
                    return self.state

        self.state = ActionStateEnum.SUCCESS
        return self.state


class ReproducibleBuildEnvironmentCheck(Check):
    """A Check ensuring all build requirements of OutputPackageBases can covered by a set of available requirements.

    Attributes
    ----------
    pkgbases: list[OutputPackageBase]
        A list of OutputPackageBase objects, that represent the pkgbases to be checked
    available_requirements: set[str]
        A set of package information strings (name, version, architecture), which are matched against the build
        requirements of pkgbases
    """

    def __init__(
        self,
        pkgbases: list[OutputPackageBase],
        available_requirements: set[str],
    ) -> None:
        """Initialize an instance of ReproducibleBuildEnvironmentCheck.

        Parameters
        ----------
        pkgbases: list[OutputPackageBase]
            A list of OutputPackageBase objects, that represent the pkgbases to be checked
        available_requirements: set[str]
            A set of package information strings (name, version, architecture), which are matched against the build
            requirements of pkgbases
        """
        self.pkgbases = pkgbases
        self.available_requirements = available_requirements

    def __call__(self) -> ActionStateEnum:
        """Check all build requirements of OutputPackageBases can be met by a set of available requirements.

        Returns
        -------
        ActionStateEnum
            ActionStateEnum.SUCCESS if the check passed successfully,
            ActionStateEnum.FAILED otherwise
        """
        self.state = ActionStateEnum.STARTED

        missing_requirements: dict[str, list[str]] = defaultdict(list)

        for pkgbase in self.pkgbases:
            for requirement in pkgbase.buildinfo.installed:  # type: ignore[attr-defined]
                if requirement not in self.available_requirements:
                    missing_requirements[pkgbase.base].append(requirement)  # type: ignore[attr-defined]

        self.state = ActionStateEnum.SUCCESS
        if missing_requirements:
            self.state = ActionStateEnum.FAILED
            key_value_list = [f"\n{key} => {', '.join(value)}" for key, value in missing_requirements.items()]
            info(
                "The following build requirements can not be met with this transaction, existing packages "
                "in the repositories, or the archive:"
                f"{''.join(key_value_list)}"
            )

        return self.state


class UniqueInRepoGroupCheck(Check):
    """A Check ensuring unique lists of pkgbase and package names among a group of management repository directories.

    Attributes
    ----------
    pkgbase_names: list[str]
        A list of pkgbases to look for in the management repository directories contained in repo_management_dirs
    package_names: list[str]
        A list of packages to look for in the management repository directories contained in repo_management_dirs
    repo_management_dirs: dict[Path, list[Path]]
        A dict describing a repository name and all of its management repository directories
    """

    def __init__(
        self,
        pkgbase_names: list[str],
        package_names: list[str],
        repo_management_dirs: dict[Path, list[Path]],
    ) -> None:
        """Initialize an instance of UniqueInRepoGroupCheck.

        Parameters
        ----------
        pkgbase_names: list[str]
            A list of pkgbases to look for in the management repository directories contained in repo_management_dirs
        package_names: list[str]
            A list of packages to look for in the management repository directories contained in repo_management_dirs
        repo_management_dirs: dict[Path, list[Path]]
            A dict describing a repository name and all of its management repository directories
        """
        self.pkgbase_names = pkgbase_names
        self.package_names = package_names
        self.repo_management_dirs = repo_management_dirs

    def __call__(self) -> ActionStateEnum:
        """Check, that lists of pkgbase and package names are unique among a group of management repository directories.

        Returns
        -------
        ActionStateEnum
            ActionStateEnum.SUCCESS if the check passed successfully,
            ActionStateEnum.FAILED otherwise
        """
        self.state = ActionStateEnum.STARTED

        existing_pkgbases: dict[str, list[str]] = defaultdict(list)
        existing_packages: dict[str, list[str]] = defaultdict(list)

        for repo, management_dirs in self.repo_management_dirs.items():
            for management_dir in management_dirs:
                for pkgbase in self.pkgbase_names:
                    file = management_dir / f"{pkgbase}.json"
                    tmp_file = management_dir / f"{pkgbase}.json.tmp"
                    if file.exists() or tmp_file.exists():
                        existing_pkgbases[str(repo)].append(pkgbase)

                for package in self.package_names:
                    file = management_dir / "pkgnames" / f"{package}.json"
                    tmp_file = management_dir / "pkgnames" / f"{package}.json.tmp"
                    if file.exists() or tmp_file.exists():
                        existing_packages[str(repo)].append(package)

        self.state = ActionStateEnum.SUCCESS
        if existing_pkgbases or existing_packages:
            self.state = ActionStateEnum.FAILED
            pkgbases = [
                f"\nRepository '{key}' already contains the pkgbase(s) {', '.join(value)}."
                for key, value in existing_pkgbases.items()
            ]
            packages = [
                f"\nRepository '{key}' already contains the package(s) {', '.join(value)}."
                for key, value in existing_pkgbases.items()
            ]
            info(
                "Duplicate pkgbases and/ or packages detected in repository group:"
                f"{''.join(pkgbases)}"
                f"{''.join(packages)}"
            )

        return self.state



"""Handling of repod management repositories."""

from repod.repo.management.outputpackage import (  # noqa: F401
    Files,
    OutputBuildInfo,
    OutputPackage,
    OutputPackageBase,
    PackageDesc,
    export_schemas,
)



"""Reading and writing of OutputPackageBase and OutputPackage."""

from __future__ import annotations

from logging import debug
from pathlib import Path
from typing import Any

from aiofiles import open as async_open
from orjson import (
    OPT_APPEND_NEWLINE,
    OPT_INDENT_2,
    OPT_SORT_KEYS,
    JSONDecodeError,
    dumps,
    loads,
)
from pydantic import BaseModel, SerializeAsAny, ValidationError

from repod import errors
from repod.common.enums import (
    FilesVersionEnum,
    OutputPackageVersionEnum,
    PackageDescVersionEnum,
)
from repod.common.models import (
    Arch,
    Backup,
    Base,
    BuildDate,
    CheckDepends,
    Conflicts,
    CSize,
    Depends,
    Desc,
    FileName,
    Groups,
    ISize,
    License,
    MakeDepends,
    Md5Sum,
    Name,
    OptDepends,
    Packager,
    PgpSig,
    Provides,
    Replaces,
    SchemaVersionV1,
    SchemaVersionV2,
    Sha256Sum,
    SourceUrl,
    Url,
    Version,
)
from repod.files import package
from repod.files.buildinfo import (
    BuildDir,
    BuildEnv,
    BuildInfo,
    BuildInfoV1,
    BuildInfoV2,
    BuildTool,
    BuildToolVer,
    Installed,
    Options,
    PkgBuildSha256Sum,
    StartDir,
)
from repod.repo.package.syncdb import (
    Files,
    FilesV1,
    PackageDesc,
    PackageDescV1,
    PackageDescV2,
)

OUTPUT_PACKAGE_VERSIONS: dict[int, dict[str, set[str]]] = {
    1: {
        "required": {
            "arch",
            "builddate",
            "csize",
            "desc",
            "filename",
            "isize",
            "license",
            "md5sum",
            "name",
            "sha256sum",
            "url",
        },
        "optional": {
            "backup",
            "checkdepends",
            "conflicts",
            "depends",
            "files",
            "groups",
            "optdepends",
            "pgpsig",
            "provides",
            "replaces",
        },
    },
    2: {
        "required": {
            "arch",
            "builddate",
            "csize",
            "desc",
            "filename",
            "isize",
            "license",
            "name",
            "sha256sum",
            "url",
        },
        "optional": {
            "backup",
            "checkdepends",
            "conflicts",
            "depends",
            "files",
            "groups",
            "optdepends",
            "pgpsig",
            "provides",
            "replaces",
        },
    },
}
OUTPUT_PACKAGE_BASE_VERSIONS: dict[int, dict[str, int | set[str]]] = {
    1: {
        "required": {
            "base",
            "makedepends",
            "packager",
            "packages",
            "version",
        },
    },
}
DEFAULT_OUTPUT_PACKAGE_BASE_VERSION = 1


class OutputBuildInfo(BaseModel):
    """A class tracking BuildInfo information of packages that are added to instances of OutputPackageBase.

    This class is a base template class and should not be used directly.
    Instead, instantiate one of its versioned child classes using the `from_buildinfo()` classmethod.
    """

    @classmethod
    def from_buildinfo(cls, buildinfo: BuildInfo) -> OutputBuildInfo:
        """Create and return an instance of one of OutputBuildInfo's versioned child classes.

        Parameters
        ----------
        buildinfo: BuildInfo
            An instance of BuildInfo which will be used to create an instance of one of the versioned child classes of
            OutputBuildInfo

        Raises
        ------
        RuntimeError
            If an unknown BuildInfo instance is encountered

        Returns
        -------
        OutputBuildInfo
            One of OutputBuildInfo's versioned child classes (e.g. OutputBuildInfoV1 or OutputBuildInfoV2)
        """
        if isinstance(buildinfo, BuildInfoV1):
            return OutputBuildInfoV1(
                builddir=buildinfo.builddir,
                buildenv=buildinfo.buildenv,
                installed=buildinfo.installed,
                options=buildinfo.options,
                pkgbuild_sha256sum=buildinfo.pkgbuild_sha256sum,
                schema_version=buildinfo.schema_version,
            )
        elif isinstance(buildinfo, BuildInfoV2):
            return OutputBuildInfoV2(
                builddir=buildinfo.builddir,
                buildenv=buildinfo.buildenv,
                buildtool=buildinfo.buildtool,
                buildtoolver=buildinfo.buildtoolver,
                installed=buildinfo.installed,
                options=buildinfo.options,
                pkgbuild_sha256sum=buildinfo.pkgbuild_sha256sum,
                schema_version=buildinfo.schema_version,
                startdir=buildinfo.startdir,
            )
        else:
            raise RuntimeError(
                "An unknown input format has been encountered while transforming a package's BuildInfo information "
                "into the respective output format!"
            )


class OutputBuildInfoV1(
    BuildDir,
    BuildEnv,
    Installed,
    Options,
    OutputBuildInfo,
    PkgBuildSha256Sum,
    SchemaVersionV1,
):
    """OutputBuildInfo version 1.

    Attributes which are already covered by OutputPackageBase are ommitted.
    Instances of this class relate to OutputBuildInfo the same way as BuildinfoV1 relates to BuildInfo.

    Attributes
    ----------
    builddir: str
        A string representing an absolute directory
    buildenv: list[str]
        A list of strings as described by makepkg.conf's BUILDENV option
    installed: list[str]
        A list of strings representing <package_name>-<epoch><version>-<pkgrel>-<architecture> of packages installed
        during the creation of a package
    options: list[str]
        A list of strings representing makepkg.conf OPTIONS used during the creation of a package
    pkgbuild_sha256sum: str
        A string representing a SHA-256 checksum for a PKGBUILD of a package
    schema_version: int
        An integer describing a BuildInfo format version
    """

    pass


class OutputBuildInfoV2(
    BuildDir,
    BuildEnv,
    BuildTool,
    BuildToolVer,
    Installed,
    Options,
    OutputBuildInfo,
    PkgBuildSha256Sum,
    SchemaVersionV2,
    StartDir,
):
    """OutputBuildInfo version 2.

    Attributes which are already covered by OutputPackageBase are ommitted.
    Instances of this class relate to OutputBuildInfo the same way as BuildinfoV2 relates to BuildInfo.

    Attributes
    ----------
    builddir: str
        A string representing an absolute directory
    buildenv: list[str]
        A list of strings as described by makepkg.conf's BUILDENV option
    buildtool: str
        The package name of the build tool used to create a package
    buildtoolver: str
        The version of the build tool used to create a package
    installed: list[str]
        A list of strings representing <package_name>-<epoch><version>-<pkgrel>-<architecture> of packages installed
        during the creation of a package
    options: list[str]
        A list of strings representing makepkg.conf OPTIONS used during the creation of a package
    pkgbuild_sha256sum: str
        A string representing a SHA-256 checksum for a PKGBUILD of a package
    schema_version: int
        An integer describing a BuildInfo format version
    startdir: str
        A string representing the absolute startdir directory of a package
    """

    pass


class OutputPackage(BaseModel):
    """A template class to describe all required attributes that define a package in the context of an output file.

    This class should not be instantiated directly. Use one of its subclasses instead!
    """

    @classmethod
    def from_package(cls, package: package.Package) -> OutputPackage:
        """Create an OutputPackage from a Package.

        Parameters
        ----------
        package: Path
            The path to a package file
        signature: Path | None
            The optional path to a signature file for package

        Returns
        -------
        OutputPackage
            An instance of one of OutputPackage's child classes
        """
        outputpackage_version = 0
        data = package.top_level_dict()
        keys = set(data.keys())

        debug(f"Creating OutputPackage from Package {data.get('filename')}...")

        for version in range(len(OUTPUT_PACKAGE_VERSIONS), 0, -1):
            debug(f"Testing Package keys against OutputPackage version {version}...")
            if len(OUTPUT_PACKAGE_VERSIONS[version]["required"] - {"files"} - keys) == 0:
                debug(f"OutputPackage version {version} matches the provided Package keys!")
                outputpackage_version = version
                break

        match outputpackage_version:
            case 1:
                return OutputPackageV1(
                    arch=package.pkginfo.arch,  # type: ignore[attr-defined]
                    backup=package.pkginfo.backup,  # type: ignore[attr-defined]
                    builddate=package.pkginfo.builddate,  # type: ignore[attr-defined]
                    checkdepends=package.pkginfo.checkdepends,  # type: ignore[attr-defined]
                    conflicts=package.pkginfo.conflicts,  # type: ignore[attr-defined]
                    csize=package.csize,  # type: ignore[attr-defined]
                    depends=package.pkginfo.depends,  # type: ignore[attr-defined]
                    desc=package.pkginfo.desc,  # type: ignore[attr-defined]
                    filename=package.filename,  # type: ignore[attr-defined]
                    files=Files.from_dict(
                        {
                            "files": [
                                str(path)[1:]
                                for path in package.mtree.get_paths(show_all=False)  # type: ignore[attr-defined]
                            ],
                        }
                    ),
                    groups=package.pkginfo.groups,  # type: ignore[attr-defined]
                    isize=package.pkginfo.isize,  # type: ignore[attr-defined]
                    license=package.pkginfo.license,  # type: ignore[attr-defined]
                    md5sum=package.md5sum,  # type: ignore[attr-defined]
                    name=package.pkginfo.name,  # type: ignore[attr-defined]
                    optdepends=package.pkginfo.optdepends,  # type: ignore[attr-defined]
                    pgpsig=package.pgpsig,  # type: ignore[attr-defined]
                    provides=package.pkginfo.provides,  # type: ignore[attr-defined]
                    replaces=package.pkginfo.replaces,  # type: ignore[attr-defined]
                    sha256sum=package.sha256sum,  # type: ignore[attr-defined]
                    url=package.pkginfo.url,  # type: ignore[attr-defined]
                )
            case 2:
                return OutputPackageV2(
                    arch=package.pkginfo.arch,  # type: ignore[attr-defined]
                    backup=package.pkginfo.backup,  # type: ignore[attr-defined]
                    builddate=package.pkginfo.builddate,  # type: ignore[attr-defined]
                    checkdepends=package.pkginfo.checkdepends,  # type: ignore[attr-defined]
                    conflicts=package.pkginfo.conflicts,  # type: ignore[attr-defined]
                    csize=package.csize,  # type: ignore[attr-defined]
                    depends=package.pkginfo.depends,  # type: ignore[attr-defined]
                    desc=package.pkginfo.desc,  # type: ignore[attr-defined]
                    filename=package.filename,  # type: ignore[attr-defined]
                    files=Files.from_dict(
                        {
                            "files": [
                                str(path)[1:]
                                for path in package.mtree.get_paths(show_all=False)  # type: ignore[attr-defined]
                            ],
                        }
                    ),
                    groups=package.pkginfo.groups,  # type: ignore[attr-defined]
                    isize=package.pkginfo.isize,  # type: ignore[attr-defined]
                    license=package.pkginfo.license,  # type: ignore[attr-defined]
                    name=package.pkginfo.name,  # type: ignore[attr-defined]
                    optdepends=package.pkginfo.optdepends,  # type: ignore[attr-defined]
                    pgpsig=package.pgpsig,  # type: ignore[attr-defined]
                    provides=package.pkginfo.provides,  # type: ignore[attr-defined]
                    replaces=package.pkginfo.replaces,  # type: ignore[attr-defined]
                    sha256sum=package.sha256sum,  # type: ignore[attr-defined]
                    url=package.pkginfo.url,  # type: ignore[attr-defined]
                )
            case _:
                raise RuntimeError(
                    "An error occurred trying to create an OutputPackage from a Package! "
                    f"Unable to find matching version for Package keys: {keys}"
                )


class OutputPackageV1(
    OutputPackage,
    Arch,
    Backup,
    BuildDate,
    CheckDepends,
    Conflicts,
    CSize,
    Depends,
    Desc,
    FileName,
    Groups,
    ISize,
    License,
    Md5Sum,
    Name,
    OptDepends,
    PgpSig,
    Provides,
    Replaces,
    SchemaVersionV1,
    Sha256Sum,
    Url,
):
    """A model describing all required attributes that define a package in the context of an output file (version 1).

    Attributes
    ----------
    arch: str
        The attribute can be used to describe the (required) data below an %ARCH% identifier in a 'desc' file, which
        identifies a package's architecture
    backup: list[str] | None
        The attribute can be used to describe the (optional) data below a %BACKUP% identifier in a 'desc' file, which
        identifies which file(s) of a package pacman will create backups for
    builddate: int
        The attribute can be used to describe the (required) data below a %BUILDDATE% identifier in a 'desc' file,
        which identifies a package's build date (represented in seconds since the epoch)
    checkdepends: list[str] | None
        The attribute can be used to describe the (optional) data below a %CHECKDEPENDS% identifier in a 'desc' file,
        which identifies a package's checkdepends
    conflicts: list[str] | None
        The attribute can be used to describe the (optional) data below a %CONFLICTS% identifier in a 'desc' file, which
        identifies what other package(s) a package conflicts with
    csize: int
        The attribute can be used to describe the (required) data below a %CSIZE% identifier in a 'desc' file, which
        identifies a package's size
    depends: list[str] | None
        The attribute can be used to describe the (optional) data below a %DEPENDS% identifier in a 'desc' file, which
        identifies what other package(s) a package depends on
    desc: str
        The attribute can be used to describe the (required) data below a %DESC% identifier in a 'desc' file, which
        identifies a package's description
    filename: str
        The attribute can be used to describe the (required) data below a %FILENAME% identifier in a 'desc' file, which
        identifies a package's file name
    files: list[str] | None
        The attribute can be used to describe the (optional) data below a %FILES% identifier in a 'files' file, which
        identifies which file(s) belong to a package
    groups: list[str] | None
        The attribute can be used to describe the (optional) data below a %GROUPS% identifier in a 'desc' file, which
        identifies a package's groups
    isize: int
        The attribute can be used to describe the (required) data below an %ISIZE% identifier in a 'desc' file, which
        identifies a package's installed size
    license: list[str]
        The attribute can be used to describe the (required) data below a %LICENSE% identifier in a 'desc' file, which
        identifies a package's license(s)
    md5sum: str
        The attribute can be used to describe the (required) data below an %MD5SUM% identifier in a 'desc' file, which
        identifies a package's md5 checksum
    name: str
        The attribute can be used to describe the (required) data below a %NAME% identifier in a 'desc' file, which
        identifies a package's name
    optdepends: list[str] | None
        The attribute can be used to describe the (optional) data below a %OPTDEPENDS% identifier in a 'desc' file,
        which identifies what other package(s) a package optionally depends on
    pgpsig: str | None
        The attribute can be used to describe the (required) data below a %PGPSIG% identifier in a 'desc' file, which
        identifies a package's PGP signature
    provides: list[str] | None
        The attribute can be used to describe the (optional) data below a %PROVIDES% identifier in a 'desc' file, which
        identifies what other package(s) a package provides
    replaces: list[str] | None
        The attribute can be used to describe the (optional) data below a %REPLACES% identifier in a 'desc' file, which
        identifies what other package(s) a package replaces
    schema_version: PositiveInt
        A positive integer - 1 - identifying the schema version of the object
    sha256sum: str
        The attribute can be used to describe the (required) data below an %SHA256SUM% identifier in a 'desc' file,
        which identifies a package's sha256 checksum
    url: str
        The attribute can be used to describe the (required) data below a %URL% identifier in a 'desc' file, which
        identifies a package's URL
    """

    files: SerializeAsAny[Files | None] = None


class OutputPackageV2(
    OutputPackage,
    Arch,
    Backup,
    BuildDate,
    CheckDepends,
    Conflicts,
    CSize,
    Depends,
    Desc,
    FileName,
    Groups,
    ISize,
    License,
    Name,
    OptDepends,
    PgpSig,
    Provides,
    Replaces,
    SchemaVersionV2,
    Sha256Sum,
    Url,
):
    """A model describing all required attributes that define a package in the context of an output file (version 2).

    Attributes
    ----------
    arch: str
        The attribute can be used to describe the (required) data below an %ARCH% identifier in a 'desc' file, which
        identifies a package's architecture
    backup: list[str] | None
        The attribute can be used to describe the (optional) data below a %BACKUP% identifier in a 'desc' file, which
        identifies which file(s) of a package pacman will create backups for
    builddate: int
        The attribute can be used to describe the (required) data below a %BUILDDATE% identifier in a 'desc' file,
        which identifies a package's build date (represented in seconds since the epoch)
    checkdepends: list[str] | None
        The attribute can be used to describe the (optional) data below a %CHECKDEPENDS% identifier in a 'desc' file,
        which identifies a package's checkdepends
    conflicts: list[str] | None
        The attribute can be used to describe the (optional) data below a %CONFLICTS% identifier in a 'desc' file, which
        identifies what other package(s) a package conflicts with
    csize: int
        The attribute can be used to describe the (required) data below a %CSIZE% identifier in a 'desc' file, which
        identifies a package's size
    depends: list[str] | None
        The attribute can be used to describe the (optional) data below a %DEPENDS% identifier in a 'desc' file, which
        identifies what other package(s) a package depends on
    desc: str
        The attribute can be used to describe the (required) data below a %DESC% identifier in a 'desc' file, which
        identifies a package's description
    filename: str
        The attribute can be used to describe the (required) data below a %FILENAME% identifier in a 'desc' file, which
        identifies a package's file name
    files: list[str] | None
        The attribute can be used to describe the (optional) data below a %FILES% identifier in a 'files' file, which
        identifies which file(s) belong to a package
    groups: list[str] | None
        The attribute can be used to describe the (optional) data below a %GROUPS% identifier in a 'desc' file, which
        identifies a package's groups
    isize: int
        The attribute can be used to describe the (required) data below an %ISIZE% identifier in a 'desc' file, which
        identifies a package's installed size
    license: list[str]
        The attribute can be used to describe the (required) data below a %LICENSE% identifier in a 'desc' file, which
        identifies a package's license(s)
    md5sum: str
        The attribute can be used to describe the (required) data below an %MD5SUM% identifier in a 'desc' file, which
        identifies a package's md5 checksum
    name: str
        The attribute can be used to describe the (required) data below a %NAME% identifier in a 'desc' file, which
        identifies a package's name
    optdepends: list[str] | None
        The attribute can be used to describe the (optional) data below a %OPTDEPENDS% identifier in a 'desc' file,
        which identifies what other package(s) a package optionally depends on
    pgpsig: str | None
        The attribute can be used to describe the (required) data below a %PGPSIG% identifier in a 'desc' file, which
        identifies a package's PGP signature
    provides: list[str] | None
        The attribute can be used to describe the (optional) data below a %PROVIDES% identifier in a 'desc' file, which
        identifies what other package(s) a package provides
    replaces: list[str] | None
        The attribute can be used to describe the (optional) data below a %REPLACES% identifier in a 'desc' file, which
        identifies what other package(s) a package replaces
    schema_version: PositiveInt
        A positive integer - 1 - identifying the schema version of the object
    sha256sum: str
        The attribute can be used to describe the (required) data below an %SHA256SUM% identifier in a 'desc' file,
        which identifies a package's sha256 checksum
    url: str
        The attribute can be used to describe the (required) data below a %URL% identifier in a 'desc' file, which
        identifies a package's URL
    """

    files: SerializeAsAny[Files | None] = None


class OutputPackageBase(BaseModel):
    """A template class with helper methods to create instances of one of its (versioned) subclasses.

    This class should not be instantiated directly, as it only provides generic instance methods for its subclasses.

    NOTE: The `from_dict()` classmethod is used to create one of the versioned subclasses.
    """

    @classmethod
    def from_dict(cls, data: dict[str, Any | list[Any]]) -> OutputPackageBase:
        """Create an instance of one of OutputPackageBase's subclasses from a dict.

        This method expects data derived from reading a pkgbase JSON file from the management repository.
        By default this method will prefer to create an instances of OutputPackageBase's subclasses as identified by
        DEFAULT_OUTPUT_PACKAGE_BASE_VERSION (if it is newer than the schema_version retrieved from the data).

        Parameters
        ----------
        data: dict[str, Any | list[Any]]
            A dict containing data required to instantiate a subclass of OutputPackageBase

        Raises
        ------
        RepoManagementValidationError
            If an invalid schema_version is encountered when reading data
            If a ValidationError is encountered during instantiation of one of the OutputPackage subclasses
            If an unsupported schema_version is encountered

        Returns
        -------
        OutputPackageBase
            An instance of one of the subclasses of OutputPackageBase
        """

        def default_output_package_from_dict(version: int, package: dict[str, Any]) -> OutputPackage:
            """Create the default OutputPackage subclass for a OutputPackageBase from a dict.

            Parameters
            ----------
            version: int
                The schema_version of the OutputPackageBase subclass for which to create an instance of a subclass of
                OutputPackage
            package: dict[str, Any]
                A dict describing the attributes of an OutputPackage instance

            Returns
            -------
            OutputPackage
                A subclass of OutputPackage that is the default for the given OutputPackageBase schema_version
            """
            outputpackage_version = OutputPackageVersionEnum.DEFAULT.value
            files_version = FilesVersionEnum.DEFAULT.value

            match (outputpackage_version, files_version):
                case (1, 1):
                    files = package.get("files")
                    if files:
                        if files.get("schema_version"):
                            del files["schema_version"]
                        package["files"] = FilesV1.model_validate(files)

                    return OutputPackageV1.model_validate(package)
                case (2, 1):
                    files = package.get("files")
                    del package["schema_version"]
                    if files:
                        if files.get("schema_version"):
                            del files["schema_version"]
                        package["files"] = FilesV1.model_validate(files)

                    return OutputPackageV2.model_validate(package)
                # NOTE: the catch all can never be reached but is here to satisfy our tooling
                case _:  # pragma: no cover
                    raise RuntimeError(
                        f"Invalid version provided for OutputPackage ({outputpackage_version} "
                        f"and/ or Files ({files_version})."
                    )

        used_schema_version = data.get("schema_version")
        if isinstance(used_schema_version, int):
            if DEFAULT_OUTPUT_PACKAGE_BASE_VERSION > used_schema_version:
                used_schema_version = DEFAULT_OUTPUT_PACKAGE_BASE_VERSION
                del data["schema_version"]
            elif DEFAULT_OUTPUT_PACKAGE_BASE_VERSION < used_schema_version:
                raise errors.RepoManagementValidationError(
                    f"The unsupported schema version ({used_schema_version}) has been encountered, when "
                    f"attempting to read data:\n{data}"
                )

        match used_schema_version:
            case 1:
                if isinstance(data.get("packages"), list):
                    data["packages"] = [
                        default_output_package_from_dict(version=used_schema_version, package=package)
                        for package in data.get("packages")  # type: ignore[union-attr]
                    ]
                try:
                    return OutputPackageBaseV1.model_validate(data)
                except ValidationError as e:
                    raise errors.RepoManagementValidationError(
                        "A validation error occured while attempting to instantiate an OutputPackageBaseV1 using data:"
                        f"\n'{data}'\n{e}"
                    )
            case _:
                raise errors.RepoManagementValidationError(
                    f"The unsupported schema version ({used_schema_version}) has been encountered, when "
                    f"attempting to read data:\n{data}"
                )

    @classmethod
    async def from_file(cls, path: Path) -> OutputPackageBase:
        """Initialize an OutputPackageBase from a JSON file.

        Parameters
        ----------
        path: Path
            A Path to to a JSON file

        Raises
        ------
        RepoManagementFileError
            If the JSON file can not be decoded

        Returns
        -------
        OutputPackageBase
            An instance of OutputPackageBase based on path
        """
        async with async_open(path, "r") as input_file:
            try:
                return OutputPackageBase.from_dict(data=loads(await input_file.read()))
            except JSONDecodeError as e:
                raise errors.RepoManagementFileError(f"The JSON file '{path}' could not be decoded!\n{e}")

    @classmethod
    def from_package(cls, packages: list[package.Package]) -> OutputPackageBase:
        """Create an OutputPackageBase from a list of Packages of the same pkgbase, pkgtype and version.

        Parameters
        ----------
        packages: list[Package]
            A list of Packages to create an OutputPackageBase for

        Raises
        ------
        ValueError
            If packages is an empty list,
            if more than one pkgbase is used in the list of packages,
            if duplicate package names are found in the list of packages,
            if mismatching versions are found in the list of packages,
            or if mismatching pkgtypes are found in the list of packages

        Returns
        -------
        OutputPackageBase
            An OutputPackageBase that is based on packages
        """
        if len(packages) == 0:
            raise ValueError("At least one Package needs to be provided to create an OutputPackageBase.")

        pkgbases = set([str(pkg.top_level_dict().get("pkgbase")) for pkg in packages])
        if len(pkgbases) > 1:
            raise ValueError(
                "Only one pkgbase can be used per OutputPackageBase, but Packages with the following pkgbases are "
                f"provided: {', '.join(pkgbases)}"
            )

        names = [
            str(pkg.top_level_dict().get("name")) for pkg in packages if pkg.top_level_dict().get("name") is not None
        ]
        if len(names) != len(set(names)):
            raise ValueError(
                "An error occured creating an OutputPackageBase from Packages: "
                f"No duplicate packages are allowed, but the following Package names are provided: {', '.join(names)}"
            )

        versions = set([str(pkg.top_level_dict().get("version")) for pkg in packages])
        if len(versions) != 1:
            raise ValueError(
                "Only one version can be used per OutputPackageBase, but Packages with the following versions are "
                f"provided: {', '.join(versions)}"
            )

        all_xdata: list[list[dict[str, Any]]] = [
            pkg.top_level_dict().get("xdata")  # type: ignore[misc]
            for pkg in packages
            if pkg.top_level_dict().get("xdata")
        ]
        debug(f"all xdata: {all_xdata}")
        xdata: list[dict[str, str]] = []
        xdata = [item for sublist in all_xdata for item in sublist]
        debug(f"collected xdata: {xdata}")
        pkgtypes: list[str] = []
        pkgtypes = [data.get("pkgtype").value for data in xdata if data.get("pkgtype")]  # type: ignore[union-attr]
        debug(f"collected pkgtypes: {pkgtypes}")

        debug(f"pkgtypes: {pkgtypes}")
        if len(set(pkgtypes)) > 1:
            raise ValueError(
                "An error occurred while trying to create an OutputPackageBase from Packages: "
                "Only one pkgtype can be present in the list of used Packages, but several "
                f"({', '.join(set(pkgtypes))}) "
                "are provided!"
            )

        outputpackagebase_version = 0
        data = packages[0].top_level_dict()
        keys = set(data.keys())

        debug(f"Creating OutputPackageBase from Packages {', '.join(names)}...")
        for version in range(len(OUTPUT_PACKAGE_BASE_VERSIONS), 0, -1):
            debug(f"Testing Package keys against OutputPackageBase version {version}...")
            if (
                len(
                    OUTPUT_PACKAGE_BASE_VERSIONS[version]["required"]  # type: ignore[arg-type]
                    - {"packages"}  # type: ignore[operator]
                    - keys  # type: ignore[operator]
                )
                == 0
            ):
                debug(f"OutputPackageBase version {version} matches the provided Package keys!")
                outputpackagebase_version = version
                break

        match outputpackagebase_version:
            case 1:
                return OutputPackageBaseV1(
                    base=packages[0].pkginfo.base,  # type: ignore[attr-defined]
                    buildinfo=OutputBuildInfo.from_buildinfo(packages[0].buildinfo),  # type: ignore[attr-defined]
                    makedepends=packages[0].pkginfo.makedepends,  # type: ignore[attr-defined]
                    packager=packages[0].pkginfo.packager,  # type: ignore[attr-defined]
                    packages=[OutputPackage.from_package(package=pkg) for pkg in packages],
                    version=packages[0].pkginfo.version,  # type: ignore[attr-defined]
                )
            case _:
                raise RuntimeError(
                    "An error occurred while attempting to create an OutputPackageBase!"
                    f"Unable to find matching version for Package keys: {keys}"
                )

    def add_packages(self, packages: list[OutputPackage]) -> None:
        """Add packages to an instance of one of OutputPackageBase's subclasses.

        NOTE: This method only operates successfully if the instance of the class using it actually defines the
        `packages` field! The OutputPackageBase class does not do that!

        Parameters
        ----------
        packages: list[OutputPackage]
            A list of OutputPackage instances to add

        Raises
        ------
        RuntimeError
            If called on the OutputPackageBase template class
        """
        if hasattr(self, "packages"):
            # TODO: only add OutputPackage subclasses that are compatible with the OutputPackageBase version, else
            # attempt conversion
            self.packages += packages
        else:
            raise RuntimeError("It is not possible to add packages to the template class OutputPackageBase!")

    def get_version(self) -> str:
        """Get version of an instance of one of OutputPackage's subclasses.

        NOTE: This method only successfully returns if the instance of the class using it defines the `get_version`
        field! The OutputPackageBase class does not do that!

        Raises
        ------
        RuntimeError
            If called on the OutputPackageBase template class

        Returns
        -------
        str
            The version string of the OutputPackageBase
        """
        if hasattr(self, "version"):
            return str(self.version)
        else:
            raise RuntimeError(
                "It is not possible to return the version attribute of the template class OutputPackageBase!"
            )

    async def get_packages_as_models(
        self,
        packagedesc_version: PackageDescVersionEnum = PackageDescVersionEnum.DEFAULT,
        files_version: FilesVersionEnum = FilesVersionEnum.DEFAULT,
    ) -> list[tuple[PackageDesc, Files]]:
        """Return the list of packages as tuples of PackageDesc and Files models.

        NOTE: This method only successfully returns if the instance of the class using it defines the required fields!
        The OutputPackageBase class does not do that!

        Parameters
        ----------
        packagedesc_version: int
            The PackageDesc version to use
        files_version: int
            The Files version to use

        Raises
        ------
        RuntimeError
            If this method is called on the template class OutputPackageBase (instead of on one of its subclasses)
            If an unknown schema_version is encountered in the OutputPackageBase instance.
            If unsupported packagedesc_version or files_version are encountered.

        Returns
        -------
        list[tuple[PackageDesc, Files]]
            A list of tuples with one PackageDesc and one Files each
        """
        if not hasattr(self, "schema_version"):
            raise RuntimeError(
                "Packages and their files can not be retrieved from the templatae class OutputPackageBase!"
            )

        if self.schema_version not in OUTPUT_PACKAGE_BASE_VERSIONS.keys():
            raise RuntimeError(f"OutputPackageBase has invalid schema_version {self.schema_version}!")

        match (packagedesc_version, files_version):
            case (PackageDescVersionEnum.ONE, FilesVersionEnum.DEFAULT):
                return [
                    (
                        PackageDescV1(
                            arch=package.arch,
                            backup=package.backup,
                            base=self.base,  # type: ignore[attr-defined]
                            builddate=package.builddate,
                            checkdepends=package.checkdepends,
                            conflicts=package.conflicts,
                            csize=package.csize,
                            depends=package.depends,
                            desc=package.desc,
                            filename=package.filename,
                            groups=package.groups,
                            isize=package.isize,
                            license=package.license,
                            makedepends=self.makedepends,  # type: ignore[attr-defined]
                            md5sum=package.md5sum,
                            name=package.name,
                            optdepends=package.optdepends,
                            packager=self.packager,  # type: ignore[attr-defined]
                            pgpsig=package.pgpsig,
                            provides=package.provides,
                            replaces=package.replaces,
                            sha256sum=package.sha256sum,
                            url=package.url,
                            version=self.version,  # type: ignore[attr-defined]
                        ),
                        FilesV1(files=package.files.files if package.files else []),
                    )
                    for package in self.packages  # type: ignore[attr-defined]
                ]
            case (PackageDescVersionEnum.TWO, FilesVersionEnum.DEFAULT):
                return [
                    (
                        PackageDescV2(
                            arch=package.arch,
                            backup=package.backup,
                            base=self.base,  # type: ignore[attr-defined]
                            builddate=package.builddate,
                            checkdepends=package.checkdepends,
                            conflicts=package.conflicts,
                            csize=package.csize,
                            depends=package.depends,
                            desc=package.desc,
                            filename=package.filename,
                            groups=package.groups,
                            isize=package.isize,
                            license=package.license,
                            makedepends=self.makedepends,  # type: ignore[attr-defined]
                            name=package.name,
                            optdepends=package.optdepends,
                            packager=self.packager,  # type: ignore[attr-defined]
                            provides=package.provides,
                            replaces=package.replaces,
                            sha256sum=package.sha256sum,
                            url=package.url,
                            version=self.version,  # type: ignore[attr-defined]
                        ),
                        FilesV1(files=package.files.files if package.files else []),
                    )
                    for package in self.packages  # type: ignore[attr-defined]
                ]
            # NOTE: we can never reach this unless we patch PackageDescVersionEnum and/ or FilesVersionEnum
            case _:  # pragma: no cover
                raise RuntimeError(
                    f"Invalid versions provided for Files ({files_version.value}) and/ or "
                    f"PackageDesc ({packagedesc_version.value}) provided!"
                )


class OutputPackageBaseV1(
    OutputPackageBase,
    Base,
    MakeDepends,
    Packager,
    SchemaVersionV1,
    SourceUrl,
    Version,
):
    """A model for an output file format, that describes a list of packages based upon a pkgbase (version 1).

    Attributes
    ----------
    base: str
        The attribute can be used to describe the (required) data below a %BASE% identifier in a 'desc' file, which
        identifies a package's pkgbase
    buildinfo: OutputBuildInfo | None
        An optional OutputBuildInfo, which describes the build circumstances of the OutputPackageBase. The data is not
        covered in a repository sync database and therefore optional.
    makedepends: list[str] | None
        The attribute can be used to describe the (optional) data below a %MAKEDEPENDS% identifier in a 'desc' file,
        which identifies a package's makedepends
    packager: str
        The attribute can be used to describe the (required) data below a %PACKAGER% identifier in a 'desc' file, which
        identifies a package's packager
    packages: list[OutputPackage]
        A list of OutputPackage instances that belong to the pkgbase identified by base
    schema_version: PositiveInt
        A positive integer - 1 - identifying the schema version of the object
    source_url: HttpUrl | None
        An optional url that points at sources (defaults to None)
    version: str
        The attribute can be used to describe the (required) data below a %VERSION% identifier in a 'desc' file, which
        identifies a package's version (this is the accumulation of epoch, pkgver and pkgrel)
    """

    buildinfo: SerializeAsAny[OutputBuildInfo | None] = None
    packages: SerializeAsAny[list[OutputPackage]]


def export_schemas(output: Path | str) -> None:
    """Export the JSON schema of selected pydantic models to an output directory.

    Parameters
    ----------
    output: Path
        A path to which to output the JSON schema files

    Raises
    ------
    RuntimeError
        If output is not an existing directory
    """
    classes = [OutputBuildInfoV1, OutputBuildInfoV2, OutputPackageV1, OutputPackageV2, OutputPackageBaseV1]

    if isinstance(output, str):
        output = Path(output)

    if not output.exists():
        raise RuntimeError(f"The output directory {output} must exist!")

    for class_ in classes:
        with open(output / f"{class_.__name__}.json", "wb") as f:
            f.write(
                dumps(
                    class_.model_json_schema(),  # type: ignore[attr-defined]
                    option=OPT_INDENT_2 | OPT_APPEND_NEWLINE | OPT_SORT_KEYS,
                )
            )



"""Functionality for package archiving."""

from __future__ import annotations

from pathlib import Path
from shutil import copy2

from pydantic import BaseModel, field_validator

from repod.errors import RepoManagementValidationError
from repod.repo.package.repofile import filename_parts


class CopySourceDestination(BaseModel):
    """A model describing a source and a destination path.

    Attributes
    ----------
    source: Path
        An absolute source path
    destination: Path
        An absolute destination path
    """

    source: Path
    destination: Path

    @field_validator("source", "destination")
    @classmethod
    def validate_paths(cls, path: Path) -> Path:  # noqa: N805
        """Validate the source and destination paths to ensure they are absolute.

        Parameters
        ----------
        path: Path
            A path to validate

        Raises
        ------
        ValueError
            If path is not absolute

        Returns
        -------
        Path
            A validated Path
        """
        if not path.is_absolute():
            raise ValueError(f"The provided path is not absolute: {str(path)}")
        return path

    @classmethod
    def from_archive_dir(cls, source: Path, output_dir: Path) -> CopySourceDestination:
        """Create a CopySourceDestination with an archive directory as destination.

        The destination is constructed in a directory structure below output_dir, which reflects the first letter of
        source.name, the package name of source_name and source.name.

        Parameters
        ----------
        source: Path
            The source path
        output_dir: Path
            A directory below which the destination is created

        Raises
        ------
        RepoManagementValidationError
            If source is not a Path
            or if source can not be disassembled into its filename parts

        Returns
        -------
        CopySourceDestination
            An instance of CopySourceDestination
        """
        if not isinstance(source, Path):
            raise RepoManagementValidationError(f"The provided source {source} is not a Path!")
        if not isinstance(output_dir, Path):
            raise RepoManagementValidationError(f"The provided output_dir {output_dir} is not a Path!")

        try:
            parts = filename_parts(file=source)
        except ValueError as e:
            raise RepoManagementValidationError(e) from e

        return CopySourceDestination(
            source=source, destination=output_dir / parts["name"][0] / parts["name"] / source.name
        )

    def copy_file(self) -> None:
        """Copy the file from source to destination.

        The required destination directory structure is created automatically.
        """
        self.destination.parent.mkdir(mode=int("0755", base=8), parents=True, exist_ok=True)
        copy2(src=self.source, dst=self.destination)

    def remove_destination(self) -> None:
        """Remove the destination file.

        Previously created directories are left untouched.
        """
        self.destination.unlink(missing_ok=True)



"""Handling of .MTREE files."""

from __future__ import annotations

import io
import re
from logging import debug
from pathlib import Path

from orjson import OPT_APPEND_NEWLINE, OPT_INDENT_2, OPT_SORT_KEYS, dumps
from pydantic import (
    BaseModel,
    Field,
    NonNegativeFloat,
    NonNegativeInt,
    StringConstraints,
    ValidationError,
)
from typing_extensions import Annotated

from repod.common.models import SchemaVersionV1
from repod.common.regex import ABSOLUTE_MTREE_PATH, MD5, RELATIVE_MTREE_PATH, SHA256
from repod.errors import RepoManagementValidationError


class SystemGID(BaseModel):
    """The group ID of a system group.

    Attributes
    ----------
    gid: int
        A group ID >0, <1000
    """

    gid: Annotated[int, Field(strict=True, ge=0, lt=1000)]


class LinkTarget(BaseModel):
    """The target location of a symlink.

    Attributes
    ----------
    link: str | None
        An optional string representing a relative or absolute file
    """

    link: Annotated[str, StringConstraints(pattern=rf"^({RELATIVE_MTREE_PATH}|{ABSOLUTE_MTREE_PATH})$")] | None = None


class Md5(BaseModel):
    """An MD5 checksum.

    Attributes
    ----------
    md5: str | None
        An optional string representing an MD5 checksum
    """

    md5: Annotated[str, StringConstraints(pattern=rf"^{MD5}$")] | None = None


class FileMode(BaseModel):
    """A numeric unix file mode.

    Attributes
    ----------
    mode: str
        A three or four digit long string, consisting only of valid file modes
    """

    mode: Annotated[str, StringConstraints(pattern=r"^[01234567]{3,4}$")]


class MTreeEntryName(BaseModel):
    """A file name in mtree representation.

    The mtree format allows for encoding characters using a block of backslash and three octal digits (see
    https://man.archlinux.org/man/mtree.5#General_Format).

    Attributes
    ----------
    name: str
        A string representing an absolute file location in mtree format
    """

    name: Annotated[str, StringConstraints(pattern=rf"^{ABSOLUTE_MTREE_PATH}$")]


class Sha256(BaseModel):
    """A SHA-256 checksum.

    Attributes
    ----------
    sha256:
        An optional string representing a SHA-256 checksum
    """

    sha256: Annotated[str, StringConstraints(pattern=rf"^{SHA256}$")] | None = None


class FileSize(BaseModel):
    """A file size in bytes.

    Attributes
    ----------
    size: int
        A non-negative integer describing a file size in bytes
    """

    size: NonNegativeInt | None = None


class UnixTime(BaseModel):
    """A timestamp in seconds since the epoch.

    Attributes
    ----------
    time: float
        A float > 0 representing a unix timestamp (seconds since the epoch)
    """

    time: NonNegativeFloat


class MTreeEntryType(BaseModel):
    """A file type for mtree entries (see https://man.archlinux.org/man/mtree.5#Keywords).

    Attributes
    ----------
    type_: str
        A string representing a valid mtree type (one of block, char, dir, fifo, file, link or socket)
    """

    type_: Annotated[str, StringConstraints(pattern=r"^(block|char|dir|fifo|file|link|socket)$")]


class SystemUID(BaseModel):
    """The user ID of a system user.

    Attributes
    ----------
    uid: int
        A user ID >0, <1000
    """

    uid: Annotated[int, Field(strict=True, ge=0, lt=1000)]


class MTreeEntry(BaseModel):
    """An entry in an MTree.

    This is a template class and should not be used directly. Instead instantiate one of the classes derived from it.
    """

    def get_file_path(self) -> Path:
        """Return the file name as a Path.

        The mtree format allows for encoding characters using a block of backslash and three octal digits (see
        https://man.archlinux.org/man/mtree.5#General_Format).
        This method finds and replaces occurences of these encoded characters.

        Raises
        ------
        RuntimeError
            If called on the template class MTreeEntry

        Returns
        -------
        Path
            The Path representation of name
        """
        if not hasattr(self, "name"):
            raise RuntimeError("It is not possible to retrieve a file path from the template class MTreeEntry!")

        output_name = self.name

        if len(re.findall(r"\\[0-9A-F]{3}", output_name)) > 0:
            output_name = output_name.encode("latin1").decode("unicode-escape").encode("latin1").decode("utf8")
            debug(f"Converted MTree path {self.name} to {output_name}.")

        return Path(output_name)

    def get_link_path(self, resolve: bool = False) -> Path | None:
        """Return the link as a Path.

        The mtree format allows for encoding characters using a block of backslash and three octal digits (see
        https://man.archlinux.org/man/mtree.5#General_Format).
        This method finds and replaces occurences of these encoded characters.
        Any relative paths are converted to absolute ones.

        Parameters
        ----------
        resolve: bool
            Whether to fully resolve the link - defaults to False

        Raises
        ------
        RuntimeError
            If called on the template class MTreeEntry

        Returns
        -------
        Path | None
            The Path representation of link, or None if there is None
        """
        if not hasattr(self, "link"):
            raise RuntimeError("It is not possible to retrieve a file path from the template class MTreeEntry!")

        output_name = self.link
        if output_name is None:
            return output_name

        if len(re.findall(r"\\[0-9A-F]{3}", output_name)) > 0:
            output_name = output_name.encode("latin1").decode("unicode-escape").encode("latin1").decode("utf8")
            debug(f"Converted MTree path {self.name} to {output_name}.")  # type: ignore[attr-defined]

        output_path = Path(output_name)

        if not resolve or output_path.is_absolute():
            return output_path
        else:
            if output_name.startswith(".."):
                return (self.get_file_path() / output_path).resolve()
            else:
                return (self.get_file_path() / ".." / output_path).resolve()

    def get_type(self) -> str:
        """Return the type as a string.

        Raises
        ------
        RuntimeError
            If called on the template class MTreeEntry

        Returns
        -------
        str
            The type of the MTreeEntry
        """
        if not hasattr(self, "type_"):
            raise RuntimeError("It is not possible to retrieve a type from the template class MTreeEntry!")

        return str(self.type_)


class MTreeEntryV1(
    MTreeEntry,
    SchemaVersionV1,
    FileMode,
    FileSize,
    LinkTarget,
    Md5,
    MTreeEntryName,
    MTreeEntryType,
    Sha256,
    UnixTime,
    SystemGID,
    SystemUID,
):
    """An entry in an MTree (version 1).

    Attributes
    ----------
    gid: int
        A group ID >0, <1000
    link: str | None
        An optional string representing a relative or absolute file
    md5: str | None
        A optional string representing an MD5 checksum
    mode: str
        A three or four digit long string, consisting only of valid file modes
    name: str
        A string representing an absolute file location in mtree format
    schema_version: int
        A schema version (defaults to 1)
    sha256:
        An optional string representing a SHA-256 checksum
    size: int
        A non-negative integer describing a file size in bytes
    time: float
        A float > 0 representing a unix timestamp (seconds since the epoch)
    type_: str
        A string representing a valid mtree type (one of block, char, dir, fifo, file, link or socket)
    uid: int
        A user ID >0, <1000
    """

    pass


class MTree(BaseModel):
    """A class to describe an mtree file.

    Attributes
    ----------
    files: list[File]
        A list of File instances, representing the entries in an mtree file
    """

    entries: list[MTreeEntry]

    @classmethod
    def from_file(cls, data: io.StringIO) -> MTree:
        """Create an instance of MTree from an io.StringIO representing the contents of an mtree file.

        Parameters
        ----------
        data: io.StringIO
            A text stream representing the contents of an mtree file

        Raises
        ------
        RepoManagementValidationError
            If a ValidationError occurs during validation of the data

        Returns
        -------
        MTree
            An instance of MTree, derived from data
        """

        def sanitize_mtree_pairs(
            settings_list: list[list[str]],
            settings_dict: dict[str, float | int | str],
        ) -> None:
            """Sanitize mtree pairs in a list and add them to a dict.

            Parameters
            ----------
            settings_list: list[list[str]]
                A list of string lists, that represent mtree key-value pairs
            settings_dict: dict[str, float | int | str]
                A dict to which sanitized mtree key-value pairs are added
            """
            for setting in settings_list:
                settings_key = setting[0]
                setting_value = setting[1].strip("\n")
                match settings_key:
                    case "gid" | "uid" | "size":
                        settings_dict[settings_key] = int(setting_value)
                    case "time":
                        settings_dict[settings_key] = float(setting_value)
                    case "type":
                        # NOTE: do not overload type()
                        settings_dict["type_"] = setting_value
                    case "mode":
                        # NOTE: ensure that file modes are zero-filled and of length 4
                        settings_dict[settings_key] = setting_value.zfill(4)
                    case "md5digest" | "sha1digest" | "sha256digest" | "sha384digest" | "sha512digest":
                        # NOTE: only use the name of the digest as key to be more flexible in reusing the model
                        settings_dict[settings_key.replace("digest", "")] = setting_value
                    case _:
                        settings_dict[settings_key] = setting_value

        base_settings: dict[str, float | int | str] = {}
        entries: list[MTreeEntry] = []

        for line in data:
            if line.startswith("/set"):
                settings_list = [assignment.split("=") for assignment in line.split(" ")[1:]]
                sanitize_mtree_pairs(settings_list=settings_list, settings_dict=base_settings)

            elif line.startswith("."):
                file_settings: dict[str, float | int | str] = {}
                file_settings["name"] = line.split()[0][1:]

                # provide a list of all settings in an entry line (skip empty assigments due to multiple whitespace)
                settings_list = [assignment.split("=") for assignment in line.split()[1:] if assignment]
                sanitize_mtree_pairs(settings_list=settings_list, settings_dict=file_settings)

                try:
                    entries.append(
                        MTreeEntryV1(
                            gid=file_settings.get("gid") or base_settings.get("gid"),  # type: ignore[arg-type]
                            link=file_settings.get("link"),  # type: ignore[arg-type]
                            md5=file_settings.get("md5") or base_settings.get("md5"),  # type: ignore[arg-type]
                            mode=file_settings.get("mode") or base_settings.get("mode"),  # type: ignore[arg-type]
                            name=file_settings.get("name"),  # type: ignore[arg-type]
                            sha256=file_settings.get("sha256") or base_settings.get("sha256"),  # type: ignore[arg-type]
                            size=file_settings.get("size") or base_settings.get("size"),  # type: ignore[arg-type]
                            time=file_settings.get("time") or base_settings.get("time"),  # type: ignore[arg-type]
                            type_=file_settings.get("type_") or base_settings.get("type_"),  # type: ignore[arg-type]
                            uid=file_settings.get("uid") or base_settings.get("uid"),  # type: ignore[arg-type]
                        )
                    )
                except ValidationError as e:
                    raise RepoManagementValidationError(
                        f"An error occured when validating mtree data!\n"
                        f"Basic settings: {base_settings}\n"
                        f"File settings: {file_settings}\n"
                        f"{e}"
                    )
            else:
                continue

        return MTree(entries=entries)

    def get_paths(self, show_all: bool = False) -> list[Path]:
        """Return the list of Paths described by the entries of the MTree.

        Parameters
        ----------
        show_all: bool
            Also show files that are not installed on target systems (defaults to False)

        Returns
        -------
        list[Path]
            A list of Paths
        """
        path_list: list[Path] = []

        for entry in self.entries:
            path = entry.get_file_path()
            if path in [Path("/.BUILDINFO"), Path("/.INSTALL"), Path("/.MTREE"), Path("/.PKGINFO")] and not show_all:
                continue

            path_list.append(path)

        return path_list


def export_schemas(output: Path | str) -> None:
    """Export the JSON schema of selected pydantic models to an output directory.

    Parameters
    ----------
    output: Path
        A path to which to output the JSON schema files

    Raises
    ------
    RuntimeError
        If output is not an existing directory
    """
    classes = [MTreeEntryV1]

    if isinstance(output, str):
        output = Path(output)

    if not output.exists():
        raise RuntimeError(f"The output directory {output} must exist!")

    for class_ in classes:
        with open(output / f"{class_.__name__}.json", "wb") as f:
            f.write(
                dumps(
                    class_.model_json_schema(),
                    option=OPT_INDENT_2 | OPT_APPEND_NEWLINE | OPT_SORT_KEYS,
                )
            )



"""Enums used through repod."""

from __future__ import annotations

from enum import Enum, IntEnum, IntFlag, auto


class ArchitectureEnum(Enum):
    """An Enum to distinguish different CPU architectures.

    Attributes
    ----------
    AARCH64: "aarch64"
        The aarch64 CPU architecture
    ANY: "any"
        Any CPU architecture
    ARM: "arm"
        The arm CPU architecture
    ARMV6H: "armv6h"
        The armv6h CPU architecture
    ARMV7H: "armv7h"
        The armv7h CPU architecture
    I486: "i486"
        The i486 CPU architecture
    I686: "i686"
        The i686 CPU architecture
    PENTIUM4: "pentium4"
        The pentium4 CPU architecture
    RISCV32: "riscv32"
        The risv32 CPU architecture
    RISCV64: "riscv64"
        The risv64 CPU architecture
    X86_64: "x86_64"
        The x86_64 CPU architecture
    X86_64_V2: "x86_64_v2"
        The x86_64_v2 CPU architecture
    X86_64_V3: "x86_64_v3"
        The x86_64_v3 CPU architecture
    X86_64_V4: "x86_64_v4"
        The x86_64_v4 CPU architecture
    """

    AARCH64 = "aarch64"
    ANY = "any"
    ARM = "arm"
    ARMV6H = "armv6h"
    ARMV7H = "armv7h"
    I486 = "i486"
    I686 = "i686"
    PENTIUM4 = "pentium4"
    RISCV32 = "riscv32"
    RISCV64 = "riscv64"
    X86_64 = "x86_64"
    X86_64_V2 = "x86_64_v2"
    X86_64_V3 = "x86_64_v3"
    X86_64_V4 = "x86_64_v4"

    @classmethod
    def as_or_regex(cls) -> str:
        """Return the members of ArchitectureEnum formatted as an "or" concatenated string.

        Returns
        -------
        str
            The members of ArchitectureEnum formatted as an "or" concatenated string
        """
        return r"|".join(arch.value for arch in cls)


class FieldTypeEnum(IntEnum):
    """An IntEnum to distinguish the different types of entries in a file.

    Attributes
    ----------
    STRING: int
        An entry of type 'str'
    INT: int
        An entry of typoe 'int'
    STRING_LIST: int
        An entry of type 'list[str]'
    """

    STRING = 0
    INT = 1
    STRING_LIST = 2
    KEY_VALUE_LIST = 3


class CompressionTypeEnum(Enum):
    """An Enum to distinguish different compression types of a tar file.

    The member values represents the name of possible file suffixes (without leading dot)

    Attributes
    ----------
    NONE: ""
        No compression
    BZIP2: "bz2"
        The bzip2 compression
    GZIP: "gz"
        The gzip compression
    LZMA: "xz"
        The lzma compression
    ZSTANDARD: "zst"
        The zstandard compression
    """

    NONE = ""
    BZIP2 = "bz2"
    GZIP = "gz"
    LZMA = "xz"
    ZSTANDARD = "zst"

    @classmethod
    def from_string(cls, input_: str) -> CompressionTypeEnum:
        """Return a CompressionTypeEnum member based on an input string.

        Parameters
        ----------
        input_: str
            A string representing one of the CompressionTypeEnum members. Valid options are "none", "bzip2", "bz2",
            "gzip", "gz", "lzma", "xz", "zstandard" and "zst"

        Raises
        ------
        RuntimeError
            If an invalid input is provided

        Returns
        -------
        CompressionTypeEnum
            A CompressionTypeEnum member that matches input_
        """
        match input_:
            case "none":
                return CompressionTypeEnum.NONE
            case "bzip2" | "bz2":
                return CompressionTypeEnum.BZIP2
            case "gzip" | "gz":
                return CompressionTypeEnum.GZIP
            case "lzma" | "xz":
                return CompressionTypeEnum.LZMA
            case "zstandard" | "zst":
                return CompressionTypeEnum.ZSTANDARD
            case _:
                raise RuntimeError(f"The provided compression type {input_} is not valid!")

    @classmethod
    def db_tar_suffix(cls, compression_type: CompressionTypeEnum, files: bool = False) -> str:
        """Return a member formated as the file suffix for a default or files repository sync database.

        Parameters
        ----------
        compression_type: CompressionTypeEnum
            A member of CompressionTypeEnum to return the tar suffix for
        files: bool
            Whether to return the tar suffix for a files database (defaults to False)

        Returns
        -------
        str
            A member of CompressionTypeEnum formated as the file suffix for a default repository sync database
        """
        return (
            f".{'files' if files else 'db'}.tar." + compression_type.value
            if len(compression_type.value) > 0
            else f".{'files' if files else 'db'}.tar" + compression_type.value
        )

    @classmethod
    def as_db_file_suffixes(cls) -> list[str]:
        """Return members as list of strings representing all permutations for a default repository sync database.

        Returns
        -------
        list[str]
            A list of strings representing all possible permutations of file suffixes for a default repository sync
            database
        """
        return [".db", ".db.tar"] + [".db.tar." + name.value for name in cls if len(name.value) > 0]

    @classmethod
    def as_files_file_suffixes(cls) -> list[str]:
        """Return members as list of strings representing all permutations for a files repository sync database.

        Returns
        -------
        list[str]
            A list of strings representing all possible permutations of file suffixes for a files repository sync
            database
        """
        return [".files", ".files.tar"] + [".files.tar." + name.value for name in cls if len(name.value) > 0]


class FilesVersionEnum(IntEnum):
    """An IntEnum to distinguish different version of Files.

    Attributes
    ----------
    DEFAULT: int
        The default Files version
    ONE: int
        The first Files version
    """

    DEFAULT = 1
    ONE = 1


class OutputPackageVersionEnum(IntEnum):
    """An IntEnum to distinguish different version of OutputPackage.

    Attributes
    ----------
    DEFAULT: int
        The default OutputPackage version
    ONE: int
        The first OutputPackage version
    """

    DEFAULT = 2
    ONE = 1
    TWO = 2


class PackageDescVersionEnum(IntEnum):
    """An IntEnum to distinguish different version of PackageDesc.

    Attributes
    ----------
    DEFAULT: int
        The default PackageDesc version
    ONE: int
        The first PackageDesc version
    TWO: int
        The second PackageDesc version
    """

    DEFAULT = 2
    ONE = 1
    TWO = 2


class PkgVerificationTypeEnum(Enum):
    """An Enum to distinguish different package signature verification implementations.

    Attributes
    ----------
    PACMANKEY: str
        An implementation based on pacman-key --verify
    """

    PACMANKEY = "pacman-key"


class PkgTypeEnum(Enum):
    """An Enum to distinguish different package types.

    The member values represents the name of a possible repod.files.pkginfo.PkgType value

    Attributes
    ----------
    PKG: str
        A default package
    DEBUG: str
        A debug package
    SOURCE: str
        A source package
    SPLIT: str
        A split package
    """

    PKG = "pkg"
    DEBUG = "debug"
    SOURCE = "src"
    SPLIT = "split"


class RepoFileEnum(IntEnum):
    """An Enum to distinguish different types of RepoFiles.

    Attributes
    ----------
    PACKAGE: int
        A package file
    PACKAGE_SIGNATURE: int
        A package signature file
    """

    PACKAGE = 0
    PACKAGE_SIGNATURE = 1


class RepoDirTypeEnum(Enum):
    """An Enum to distinguish different types of repository directories.

    Attributes
    ----------
    MANAGEMENT: str
        A management repository directory
    PACKAGE: str
        A package repository directory
    POOL: str
        A pool directory directory
    """

    MANAGEMENT = "management"
    PACKAGE = "package"
    POOL = "pool"


class RepoTypeEnum(Enum):
    """An Enum to distinguish different types of repositories.

    Attributes
    ----------
    STABLE: str
        A stable repository
    DEBUG: str
        A stable debug repository
    STAGING: str
        A staging repository
    STAGING_DEBUG: str
        A staging debug repository
    TESTING: str
        A testing repository
    TESTING_DEBUG: str
        A testing debug repository
    """

    STABLE = "stable"
    STABLE_DEBUG = "stable debug"
    STAGING = "staging"
    STAGING_DEBUG = "staging debug"
    TESTING = "testing"
    TESTING_DEBUG = "testing debug"

    @classmethod
    def from_bool(cls, debug: bool, staging: bool, testing: bool) -> RepoTypeEnum:
        """Return a member of RepoTypeEnum by providing boolean values.

        Parameters
        ----------
        debug: bool
            A boolean value indicating whether a debug repository is targeted
        staging: bool
            A boolean value indicating whether a staging repository is targeted
        testing: bool
            A boolean value indicating whether a testing repository is targeted

        Raises
        ------
        RuntimeError
            If no member of RepoTypeEnum matches the provided boolean values

        Returns
        -------
        RepoTypeEnum
            A member of RepoTypeEnum matching the provided boolean values
        """
        match debug, staging, testing:
            case [False, False, False]:
                return cls.STABLE
            case [False, True, False]:
                return cls.STAGING
            case [False, False, True]:
                return cls.TESTING
            case [True, False, False]:
                return cls.STABLE_DEBUG
            case [True, True, False]:
                return cls.STAGING_DEBUG
            case [True, False, True]:
                return cls.TESTING_DEBUG
            case _:
                raise RuntimeError(
                    f"Can not define a repository type from data: debug={debug}, staging={staging}, testing={testing}"
                )


class SettingsTypeEnum(Enum):
    """An Enum to distinguish different Settings types.

    Attributes
    ----------
    USER: str
        User Settings
    SYSTEM: str
        System Settings
    """

    USER = "user"
    SYSTEM = "system"


def tar_compression_types_for_filename_regex() -> str:
    """Return the members of CompressionTypeEnum formatted for use in the FILENAME regular expression.

    Returns
    -------
    str
        The members of CompressionTypeEnum formatted as an "or" concatenated string (including a leading empty match for
        no compression)
    """
    return r"|".join([type_.value for type_ in CompressionTypeEnum]).replace("|", r"|\.")


class ActionStateEnum(IntFlag):
    """An Enum to distinguish different states in Checks and Tasks.

    Attributes
    ----------
    NOT_STARTED: int
        An action is not started
    STARTED: int
        An action is started
    STARTED_TASK: int
        An action is started and is a Task
    FAILED: int
        An action is failed
    FAILED_DEPENDENCY: int
        An action's dependency is failed
    FAILED_PRE_CHECK: int
        An action is failed and is a pre Check
    FAILED_POST_CHECK: int
        An action is failed and is a post Check
    FAILED_TASK: int
        An action is failed and is a Task
    FAILED_UNDO_DEPENDENCY: int
        An action is failed and is an undo Task of a dependency
    FAILED_UNDO_TASK: int
        An action is failed and is an undo Task
    SUCCESS: int
        An action is successful
    SUCCESS_TASK: int
        An action is successful and is a Task
    """

    NOT_STARTED = auto()
    STARTED = auto()
    STARTED_TASK = auto()
    FAILED = auto()
    FAILED_DEPENDENCY = auto()
    FAILED_PRE_CHECK = auto()
    FAILED_POST_CHECK = auto()
    FAILED_TASK = auto()
    FAILED_UNDO_DEPENDENCY = auto()
    FAILED_UNDO_TASK = auto()
    SUCCESS = auto()
    SUCCESS_TASK = auto()



"""Utility functions for running external commands."""

from pathlib import Path
from subprocess import PIPE, STDOUT, CalledProcessError  # nosec: B404

from subprocess_tee import CompletedProcess, run  # nosec: B404


def _print_env(env: dict[str, str] | None) -> None:
    """Print the environment variables from a dict.

    Parameters
    ----------
    env: dict[str, str] | None
        An optional dict with environment variables and their values
    """
    if env:
        for key, value in sorted(env.items()):
            print(f"{key}: {value}")


def run_command(
    cmd: str | list[str],
    env: dict[str, str] | None = None,
    debug: bool = False,
    echo: bool = False,
    quiet: bool = False,
    check: bool = False,
    cwd: str | Path | None = None,
) -> CompletedProcess:
    """Run a command.

    Parameters
    ----------
    cmd: str | list[str]
        A string or list of strings that will be passed to subprocess.run()
    env: dict[str, str] | None
        A dict of environment variables and their respective values (defaults to None)
    debug: bool
        Whether to run in debug mode, which prints environment variables and command output (defaults to False)
    echo: bool
        Whether to print the command before running it (defaults to False)
    quiet: bool
        Whether to print the output of command while running it (defaults to False)
    check: bool
        Whether to check the return code of the command, which implies raising CallecProcessError (defaults to False)
    cwd: str | Path | None
        In which directory to run the command (defaults to None, which means current working directory)

    Raises
    ------
    CalledProcessError
        If check is True and the commands return code is not 0

    Returns
    -------
    CompletedProcess
        The result of the command
    """
    if debug:
        _print_env(env)

    result = run(
        cmd,
        env=env,
        stdout=PIPE,
        stderr=STDOUT,
        echo=echo or debug,
        quiet=quiet,
        cwd=cwd,
    )
    if result.returncode != 0 and check:
        raise CalledProcessError(
            returncode=result.returncode,
            cmd=result.args,
            output=result.stdout,
            stderr=result.stderr,
        )
    return result



"""A version utility module."""


def cmp(a: int | str, b: int | str) -> int:
    """
    C style cmp function returning the compare result as int.

    Compare the two objects a and b and return an integer according to
    the outcome. The return value is negative if x < b, zero if a == b
    and strictly positive if a > b.

    Parameters
    ----------
    a: any
        First input to compare
    b: any
        Second input to compare

    Returns
    -------
    int
        -1 if 'b' is greater
         0 if both are equal
         1 if 'a' is greater
    """
    return int(str(a) > str(b)) - int(str(a) < str(b))



"""Handling of package files and their contents."""

from __future__ import annotations

from base64 import b64encode
from hashlib import md5, sha256
from logging import debug, info
from pathlib import Path
from typing import Any

from orjson import OPT_APPEND_NEWLINE, OPT_INDENT_2, OPT_SORT_KEYS, dumps
from pydantic import BaseModel, SerializeAsAny

from repod.common.models import CSize, FileName, Md5Sum, PgpSig, Sha256Sum
from repod.errors import RepoManagementFileError
from repod.files.buildinfo import BuildInfo
from repod.files.common import extract_file_from_tarfile, names_in_tarfile, open_tarfile
from repod.files.mtree import MTree
from repod.files.pkginfo import PkgInfo

PACKAGE_VERSIONS = {
    1: {
        "required": {".BUILDINFO", ".MTREE", ".PKGINFO"},
    },
    2: {
        "required": {".BUILDINFO", ".MTREE", ".PKGINFO"},
    },
}


class Package(BaseModel):
    """Package representation.

    This is a template class and (apart from its class methods) should not be used directly. Instead instatiate one of
    the classes derived from it.
    """

    @classmethod
    async def from_file(cls, package: Path, signature: Path | None = None) -> Package:
        """Create a Package from a package file and an optional signature.

        Parameters
        ----------
        package: Path
            The path to a package file
        signature: Path | None
            The optional path to a signature file for package

        Raises
        ------
        RepoManagementFileError
            If the signature file does not match the package file.
            If the signature file does not exist.

        Returns
        -------
        Package
            A Package representing the metadata contained in package and the optional signature file
        """
        package_version = 0
        package_md5sum = ""
        package_sha256sum = ""
        pgpsig: str | None = None

        if signature:
            debug(f"Opening signature file {signature} for reading...")
            if not Path(str(package) + ".sig") == signature:
                raise RepoManagementFileError(
                    f"The signature file for {package} should be {str(package) + '.sig'}, but {signature} is provided!"
                )
            if not signature.exists():
                raise RepoManagementFileError(f"The signature file {signature} does not exist!")

            with open(signature, "rb") as signature_file:
                pgpsig = b64encode(signature_file.read()).decode("utf-8")
                debug(f"Created pgpsig: {pgpsig}")
        else:
            info(f"No signature file for package {package} provided, commencing without...")

        debug(f"Creating checksums for package {package}...")
        with open(package, "rb") as package_file:
            package_bytes = package_file.read()
            # NOTE: MD5 sums are still part of the PackageV1 API
            package_md5sum = md5(package_bytes).hexdigest()  # nosec: B324
            package_sha256sum = sha256(package_bytes).hexdigest()

        debug(f"Opening package file {package} for reading...")
        with open_tarfile(package) as tarfile:
            for version in range(len(PACKAGE_VERSIONS), 0, -1):
                debug(f"Testing data against Package version {version}...")
                if names_in_tarfile(tarfile=tarfile, names=PACKAGE_VERSIONS[version]["required"]):
                    debug(f"Package version {version} matches provided data!")
                    package_version = version
                    break

            match package_version:
                case 1:
                    return PackageV1(
                        buildinfo=BuildInfo.from_file(
                            data=await extract_file_from_tarfile(  # type: ignore[arg-type]
                                tarfile=tarfile,
                                file=".BUILDINFO",
                                as_stringio=True,
                            )
                        ),
                        csize=package.stat().st_size,
                        filename=package.name,
                        md5sum=package_md5sum,
                        mtree=MTree.from_file(
                            data=await extract_file_from_tarfile(  # type: ignore[arg-type]
                                tarfile=tarfile,
                                file=".MTREE",
                                as_stringio=True,
                                gzip_compressed=True,
                            ),
                        ),
                        pgpsig=pgpsig,
                        pkginfo=PkgInfo.from_file(
                            data=await extract_file_from_tarfile(  # type: ignore[arg-type]
                                tarfile=tarfile,
                                file=".PKGINFO",
                                as_stringio=True,
                            ),
                        ),
                        sha256sum=package_sha256sum,
                    )
                case 2:
                    return PackageV2(
                        buildinfo=BuildInfo.from_file(
                            data=await extract_file_from_tarfile(  # type: ignore[arg-type]
                                tarfile=tarfile,
                                file=".BUILDINFO",
                                as_stringio=True,
                            )
                        ),
                        csize=package.stat().st_size,
                        filename=package.name,
                        mtree=MTree.from_file(
                            data=await extract_file_from_tarfile(  # type: ignore[arg-type]
                                tarfile=tarfile,
                                file=".MTREE",
                                as_stringio=True,
                                gzip_compressed=True,
                            ),
                        ),
                        pgpsig=pgpsig,
                        pkginfo=PkgInfo.from_file(
                            data=await extract_file_from_tarfile(  # type: ignore[arg-type]
                                tarfile=tarfile,
                                file=".PKGINFO",
                                as_stringio=True,
                            ),
                        ),
                        sha256sum=package_sha256sum,
                    )
                case _:
                    raise RepoManagementFileError(
                        f"The provided file {package} does not match any known package versions!"
                    )

    def top_level_dict(self) -> dict[str, Any]:
        """Flatten the keys and values tracked by Package (one level deep) and return them in a dict.

        NOTE: Duplicate entries are merged!

        Returns
        -------
        dict[str, Any]
            A flattened dict representation of Package
        """
        top_level: dict[str, Any] = {}
        for key, value in self.model_dump(mode="python").items():
            if isinstance(value, dict):
                top_level.update(value)
            else:
                top_level.update({key: value})

        return top_level


class PackageV1(CSize, FileName, Md5Sum, Package, PgpSig, Sha256Sum):
    """Package representation version 1.

    Attributes
    ----------
    buildinfo: BuildInfo
        A .BUILDINFO file representation
    csize: CSize
        The file size of the Package
    filename: FileName
        The filename of the Package
    md5sum: str
        An MD5 checksum for the package
    mtree: MTree
        An .MTREE file representation
    pgpsig: str | None
        An optional PGP signature (in base64 representation) for the package
    pkginfo: PkgInfo
        A .PKGINFO file representation
    sha256sum: str
        A SHA256 checksum for the package
    """

    buildinfo: SerializeAsAny[BuildInfo]
    mtree: SerializeAsAny[MTree]
    pkginfo: SerializeAsAny[PkgInfo]


class PackageV2(CSize, FileName, Package, PgpSig, Sha256Sum):
    """Package representation version 2.

    Attributes
    ----------
    buildinfo: BuildInfo
        A .BUILDINFO file representation
    csize: CSize
        The file size of the Package
    filename: FileName
        The filename of the Package
    mtree: MTree
        An .MTREE file representation
    pgpsig: str | None
        An optional PGP signature (in base64 representation) for the package
    pkginfo: PkgInfo
        A .PKGINFO file representation
    sha256sum: str
        A SHA256 checksum for the package
    """

    buildinfo: SerializeAsAny[BuildInfo]
    mtree: SerializeAsAny[MTree]
    pkginfo: SerializeAsAny[PkgInfo]


def export_schemas(output: Path | str) -> None:
    """Export the JSON schema of selected pydantic models to an output directory.

    Parameters
    ----------
    output: Path
        A path to which to output the JSON schema files

    Raises
    ------
    RuntimeError
        If output is not an existing directory
    """
    classes = [PackageV1, PackageV2]

    if isinstance(output, str):
        output = Path(output)

    if not output.exists():
        raise RuntimeError(f"The output directory {output} must exist!")

    for class_ in classes:
        with open(output / f"{class_.__name__}.json", "wb") as f:
            f.write(
                dumps(
                    class_.model_json_schema(),  # type: ignore[attr-defined]
                    option=OPT_INDENT_2 | OPT_APPEND_NEWLINE | OPT_SORT_KEYS,
                )
            )



"""Handling of management and package repository files."""

from pathlib import Path

from repod.repo.management import (  # noqa: F401
    Files,
    OutputPackage,
    OutputPackageBase,
    PackageDesc,
)
from repod.repo.management import export_schemas as management_export_schemas
from repod.repo.package import (  # noqa: F401
    RepoDbMemberTypeEnum,
    RepoDbTypeEnum,
    SyncDatabase,
)
from repod.repo.package import export_schemas as package_export_schemas
from repod.repo.package import (  # noqa: F401
    get_desc_json_field_type,
    get_desc_json_keys,
    get_desc_json_name,
    get_files_json_field_type,
    get_files_json_keys,
    get_files_json_name,
)


def export_schemas(output: Path | str) -> None:
    """Export the JSON schema files of the repod.repo package to a directory.

    Parameters
    ----------
    output: Path
        A directory to write the JSON schema files to
    """
    management_export_schemas(output=output)
    package_export_schemas(output=output)



"""Workflows describing common repository actions."""

from logging import debug
from pathlib import Path
from sys import exit, stderr

from pydantic import AnyUrl

from repod.action.task import (
    AddToArchiveTask,
    AddToRepoTask,
    CleanupRepoTask,
    ConsolidateOutputPackageBasesTask,
    CreateOutputPackageBasesTask,
    FilesToRepoDirTask,
    MoveTmpFilesTask,
    PrintOutputPackageBasesTask,
    RemoveBackupFilesTask,
    RemoveManagementRepoSymlinksTask,
    RemovePackageRepoSymlinksTask,
    RepoGroupTask,
    ReproducibleBuildEnvironmentTask,
    Task,
    WriteOutputPackageBasesToTmpFileInDirTask,
    WriteSyncDbsToTmpFilesInDirTask,
)
from repod.common.enums import (
    ActionStateEnum,
    ArchitectureEnum,
    RepoDirTypeEnum,
    RepoFileEnum,
    RepoTypeEnum,
)
from repod.config.settings import ArchiveSettings, SystemSettings, UserSettings


def exit_on_error(message: str) -> None:
    """Print a message to stderr, optionally print argparse help and exit with return code 1.

    Parameters
    ----------
    message: str
        A message to print to stderr
    """
    print(message, file=stderr)
    stderr.write(message)
    exit(1)


def add_packages_dryrun(
    settings: SystemSettings | UserSettings,
    files: list[Path],
    repo_name: Path,
    repo_architecture: ArchitectureEnum,
    debug_repo: bool,
    with_signature: bool,
    pkgbase_urls: dict[str, AnyUrl] | None,
) -> None:
    """Print output of package representation in the management repository if packages were added.

    Parameters
    ----------
    settings: SystemSettings | UserSettings
        Settings object to retrieve data about the repository from
    files: list[Path]
        The package files to add
    repo_name: Path
        The name of the repository to add packages to
    repo_architecture: ArchitectureEnum | None
        The optional architecture of the repository to add packages to
    debug_repo: bool
        A boolean value indicating whether the packages target a debug repository
    with_signature: bool
        A boolean value indicating whether the signatures of the packages are also added
    pkgbase_urls: dict[str, AnyUrl] | None
        An optional dict, providing pkgbases and their source URLs
    """
    debug(f"Adding packages in a dry-run: {files}")
    print_task = PrintOutputPackageBasesTask(
        dumps_option=settings.get_repo_management_repo(
            name=repo_name, architecture=repo_architecture
        ).json_dumps_option,
        dependencies=[
            CreateOutputPackageBasesTask(
                architecture=settings.get_repo_architecture(name=repo_name, architecture=repo_architecture),
                package_paths=files,
                with_signature=with_signature,
                debug_repo=debug_repo,
                package_verification=settings.package_verification,
                pkgbase_urls=pkgbase_urls,
            )
        ],
    )
    if print_task() != ActionStateEnum.SUCCESS:
        print_task.undo()
        exit_on_error("An error occured while trying to add packages to a repository in a dry-run!")
        return

    return


def add_packages(
    settings: SystemSettings | UserSettings,
    files: list[Path],
    repo_name: Path,
    repo_architecture: ArchitectureEnum | None,
    debug_repo: bool,
    staging_repo: bool,
    testing_repo: bool,
    with_signature: bool,
    pkgbase_urls: dict[str, AnyUrl] | None,
) -> None:
    """Add packages to a repository.

    Parameters
    ----------
    settings: SystemSettings | UserSettings
        Settings object to retrieve data about the repository from
    files: list[Path]
        The package files to add
    repo_name: Path
        The name of the repository to add packages to
    repo_architecture: ArchitectureEnum | None
        The optional architecture of the repository to add packages to
    debug_repo: bool
        A boolean value indicating whether the packages target a debug repository
    staging_repo: bool
        A boolean value indicating whether the packages target a staging repository
    testing_repo: bool
        A boolean value indicating whether the packages target a testing repository
    with_signature: bool
        A boolean value indicating whether the signatures of the packages are also added
    pkgbase_urls: dict[str, AnyUrl] | None
        An optional dict, providing pkgbases and their source URLs
    """
    debug(f"Adding packages: {files}")
    debug(f"Provided urls: {pkgbase_urls}")

    repo = settings.get_repo(name=repo_name, architecture=repo_architecture)

    add_to_repo_dependencies: list[Task] = []

    management_repo_dir = settings.get_repo_path(
        repo_dir_type=RepoDirTypeEnum.MANAGEMENT,
        name=repo_name,
        architecture=repo_architecture,
        repo_type=RepoTypeEnum.from_bool(
            debug=debug_repo,
            staging=staging_repo,
            testing=testing_repo,
        ),
    )
    package_repo_dir = settings.get_repo_path(
        repo_dir_type=RepoDirTypeEnum.PACKAGE,
        name=repo_name,
        architecture=repo_architecture,
        repo_type=RepoTypeEnum.from_bool(
            debug=debug_repo,
            staging=staging_repo,
            testing=testing_repo,
        ),
    )

    outputpackagebasestask = CreateOutputPackageBasesTask(
        architecture=settings.get_repo_architecture(name=repo_name, architecture=repo_architecture),
        package_paths=files,
        with_signature=with_signature,
        debug_repo=debug_repo,
        pkgbase_urls=pkgbase_urls,
        package_verification=settings.package_verification,
    )
    consolidateoutputpackagebases = ConsolidateOutputPackageBasesTask(
        directory=management_repo_dir,
        stability_layer_dirs=settings.get_management_repo_stability_paths(
            name=repo_name,
            architecture=repo_architecture,
            repo_type=RepoTypeEnum.from_bool(
                debug=debug_repo,
                staging=staging_repo,
                testing=testing_repo,
            ),
        ),
        url_validation_settings=repo.package_url_validation,
        dependencies=[
            outputpackagebasestask,
        ],
    )

    package_files_task = FilesToRepoDirTask(
        files=files,
        file_type=RepoFileEnum.PACKAGE,
        settings=settings,
        name=repo_name,
        architecture=repo_architecture,
        repo_type=RepoTypeEnum.from_bool(
            debug=debug_repo,
            staging=staging_repo,
            testing=testing_repo,
        ),
    )

    if repo.build_requirements_exist:
        reproduciblebuildenvironmenttask = ReproducibleBuildEnvironmentTask(
            management_directories=[management_repo_dir],
            archive_dir=repo.archiving.packages if repo.archiving else None,  # type: ignore[union-attr]
            dependencies=[
                outputpackagebasestask,
            ],
        )
        add_to_repo_dependencies.append(reproduciblebuildenvironmenttask)

    if repo.group:
        add_to_repo_dependencies.append(
            RepoGroupTask(
                repositories=settings.get_repos_by_group(group=repo.group, exclude_repo=repo),
                dependencies=[
                    outputpackagebasestask,
                ],
            )
        )

    add_to_repo_dependencies.append(
        MoveTmpFilesTask(
            dependencies=[
                consolidateoutputpackagebases,
                WriteOutputPackageBasesToTmpFileInDirTask(
                    directory=management_repo_dir,
                    dumps_option=settings.get_repo_management_repo(
                        name=repo_name, architecture=repo_architecture
                    ).json_dumps_option,
                    dependencies=[
                        outputpackagebasestask,
                    ],
                ),
            ]
        ),
    )
    add_to_repo_dependencies.append(package_files_task)

    add_to_archive_dependencies = [
        package_files_task,
    ]

    if with_signature:
        signature_files_task = FilesToRepoDirTask(
            files=[Path(str(file) + ".sig") for file in files],
            file_type=RepoFileEnum.PACKAGE_SIGNATURE,
            settings=settings,
            name=repo_name,
            architecture=repo_architecture,
            repo_type=RepoTypeEnum.from_bool(
                debug=debug_repo,
                staging=staging_repo,
                testing=testing_repo,
            ),
        )
        add_to_repo_dependencies.append(signature_files_task)
        add_to_archive_dependencies.append(signature_files_task)

    add_to_repo_dependencies.append(
        MoveTmpFilesTask(
            dependencies=[
                WriteSyncDbsToTmpFilesInDirTask(
                    compression=settings.get_repo_database_compression(name=repo_name, architecture=repo_architecture),
                    desc_version=settings.syncdb_settings.desc_version,
                    files_version=settings.syncdb_settings.files_version,
                    management_repo_dir=management_repo_dir,
                    package_repo_dir=package_repo_dir,
                ),
            ],
        ),
    )
    if isinstance(repo.archiving, ArchiveSettings):
        add_to_repo_dependencies.append(
            AddToArchiveTask(
                archive_dir=repo.archiving.packages,
                dependencies=add_to_archive_dependencies,  # type: ignore[arg-type]
            )
        )

    add_to_repo_task = AddToRepoTask(dependencies=add_to_repo_dependencies)
    if add_to_repo_task() != ActionStateEnum.SUCCESS:
        add_to_repo_task.undo()
        exit_on_error("An error occured while trying to add packages to a repository!")
        return

    cleanup_repo_task = CleanupRepoTask(
        dependencies=[
            RemovePackageRepoSymlinksTask(
                directory=package_repo_dir,
                dependencies=[consolidateoutputpackagebases],
            ),
            RemoveManagementRepoSymlinksTask(
                directory=management_repo_dir,
                dependencies=[consolidateoutputpackagebases],
            ),
            RemoveBackupFilesTask(
                dependencies=[task for task in add_to_repo_task.dependencies if isinstance(task, MoveTmpFilesTask)]
            ),
        ],
    )
    cleanup_repo_task()

    return


def write_sync_databases(
    settings: SystemSettings | UserSettings,
    repo_name: Path,
    repo_architecture: ArchitectureEnum | None,
    debug_repo: bool,
    staging_repo: bool,
    testing_repo: bool,
) -> None:
    """Write the sync databases of a repository.

    Parameters
    ----------
    settings: SystemSettings | UserSettings
        Settings object to retrieve data about the repository from
    repo_name: Path
        The name of the repository
    repo_architecture: ArchitectureEnum | None
        The optional architecture of the repository
    debug_repo: bool
        A boolean value indicating whether to target a debug repository
    staging_repo: bool
        A boolean value indicating whether to target a staging repository
    testing_repo: bool
        A boolean value indicating whether to target a testing repository
    """
    movetmpfilestask = MoveTmpFilesTask(
        dependencies=[
            WriteSyncDbsToTmpFilesInDirTask(
                compression=settings.get_repo_database_compression(name=repo_name, architecture=repo_architecture),
                desc_version=settings.syncdb_settings.desc_version,
                files_version=settings.syncdb_settings.files_version,
                management_repo_dir=settings.get_repo_path(
                    repo_dir_type=RepoDirTypeEnum.MANAGEMENT,
                    name=repo_name,
                    architecture=repo_architecture,
                    repo_type=RepoTypeEnum.from_bool(
                        debug=debug_repo,
                        staging=staging_repo,
                        testing=testing_repo,
                    ),
                ),
                package_repo_dir=settings.get_repo_path(
                    repo_dir_type=RepoDirTypeEnum.PACKAGE,
                    name=repo_name,
                    architecture=repo_architecture,
                    repo_type=RepoTypeEnum.from_bool(
                        debug=debug_repo,
                        staging=staging_repo,
                        testing=testing_repo,
                    ),
                ),
            ),
        ],
    )
    if movetmpfilestask() != ActionStateEnum.SUCCESS:
        movetmpfilestask.undo()
        exit_on_error("An error occured while trying to write a repository's sync databases!")
        return

    remove_backup_files_task = RemoveBackupFilesTask(dependencies=[movetmpfilestask])
    remove_backup_files_task()

    return



"""Tooling for managing pacman based package repositories."""

from pathlib import Path

from .files import export_schemas as files_export_schemas
from .repo import export_schemas as repo_export_schemas


def export_schemas(output: Path | str) -> None:
    """Export the JSON schema files of repod to a directory.

    Parameters
    ----------
    output: Path
        A directory to write the JSON schema files to
    """
    files_export_schemas(output=output)
    repo_export_schemas(output=output)
